# -*- coding: utf-8 -*-
"""
Base class of SSVEP datasets
"""

import os
import abc
from typing import Union, Optional, Dict, List, Tuple, Callable
from numpy import ndarray, expand_dims
import warnings
import time

import numpy as np

import random
random.seed()

from .subjectinfo import SubInfo
from ..utils.algsupport import gen_ref_sin, floor, freqs_snr, freqs_phase, sine_snr, nextpow2
from ..evaluator.baseevaluator import create_pbar

class BaseDataset(metaclass=abc.ABCMeta):
    """
    BaseDataset
    """
    def __init__(self, 
                 subjects: List[SubInfo],
                 ID: str,
                 url: str,
                 channels: List[str],
                 srate: int,
                 block_num: int,
                 trial_num: int,
                 trial_len: float,
                 stim_info: Dict[str, Union[int, List[float]]],
                 t_prestim: float,
                 t_break: float,
                 paths: Optional[Union[str, List[str]]] = None,
                 support_files: Optional[List[str]] = None,
                 path_support_file: Optional[str] = None,
                 default_t_latency: float = 0):
        """
        Parameters required for all datasets
        
        Parameters
        -------------------
        subjects: List[SubInfo]
            List of subjects (elements stores subject information)
            
        ID: str
            Unique identifier for dataset
            
        url: str
            Download url
            
        paths: Union[str, List[str]]
            Local paths for storing data of all subjects
            
        channels: list[str]
            List of channel names
            
        srate: int
            Sampling rate
            
        block_num: int
            Number of blocks

        trial_num: int
            Number of trials in each block
            
        trial_len: float
            Signal length of single trial (in second)
            
        stim_info: Dict[int, List[float]], List[float]]
            Stimulus information. the format is
            {
                stim_num: int
                freqs: List[float]
                phases: List[float]
            }
            stim_num: Stimulus number
            freqs: Stimulus Frequencies
            phases: Stimulus phases
            
        t_prestim: float
            Pre-stimulus time
            
        t_break: float
            Time required for shifting visual attention
            
        support_files: Optional[List[str]]
            List of support file names
            
        path_support_file: Optional[str]
            Local path for storing support files
            
        default_t_latency: Optional[float]
            Default latency time
        """
        if paths is None:
            paths = os.path.join(os.getcwd(),ID)
        if type(paths) is str:
            paths = [paths for _ in range(len(subjects))]
        if path_support_file is None:
            path_support_file = paths[0] # os.path.join(os.getcwd(),'benchmark_support_file')
        if len(subjects) != len(paths):
            raise ValueError('Lengths of subjects and paths are not equal. ')
        
        
        self.subjects = subjects
        self.ID = ID
        self.url = url
        self.channels = [ch.upper() for ch in channels]
        self.srate = srate
        self.block_num = block_num
        self.trial_num = trial_num
        self.stim_info = stim_info
        self.trial_len = trial_len
        self.support_files = support_files
        self.path_support_file = path_support_file
        self.default_t_latency = default_t_latency
        self.t_prestim = t_prestim
        self.t_break = t_break
        
        # Set paths for subjects
        for sub_idx, path in enumerate(paths):
            subjects[sub_idx].path = path
        
        # download dataset
        self.download_all()
        
        # download support files
        self.download_support_files()
        
        # regist default preprocess and filterbank functions
        self.reset_preprocess()
        self.reset_filterbank()
        self.reset_ref_sig_fun()

    def default_ref_sig_fun(self, dataself, sig_len: float, N: int, phases: List[float]):
            L = floor(sig_len * dataself.srate)
            ref_sig = [gen_ref_sin(freq, dataself.srate, L, N, phase) for freq, phase in zip(dataself.stim_info['freqs'], phases)]
            return ref_sig
    def reset_ref_sig_fun(self):
        self.regist_ref_sig_fun(self.default_ref_sig_fun)
    def default_preprocess(self, dataself, X: ndarray) -> ndarray:
        """
        default preprocess (do nothing)
        """
        return X
    def reset_preprocess(self):
        self.regist_preprocess(self.default_preprocess)
    def default_filterbank(self, dataself, X: ndarray) -> ndarray:
        """
        default filterbank (1 filterbank contains original signal)
        """
        if len(X.shape) == 2:
            return expand_dims(X,0)
        elif len(X.shape) == 3:
            return X
        else:
            raise ValueError("The shapes of EEG signals are not correct")
    def reset_filterbank(self):
        self.regist_filterbank(self.default_filterbank)
    def download_all(self, total_retry_time = 10):
        """
        Download all subjects' data
        """
        for subject in self.subjects:
            download_try_count = 0
            download_flag = False
            while (not download_flag) and (download_try_count < total_retry_time):
                if download_try_count>0:
                    warnings.warn("There is an error when donwloading '{:s}' data. So retry ({:n}) after 10 seconds.".format(subject.ID, download_try_count))
                    time.sleep(10)
                download_flag_store, source_url_store, desertation_store = self.download_single_subject(subject)
                if type(download_flag_store) is not list:
                    download_flag_store = [download_flag_store]
                    source_url_store = [source_url_store]
                    desertation_store = [desertation_store]
                for download_flag, source_url, desertation in zip(download_flag_store, source_url_store, desertation_store):
                    if (not download_flag):
                        if os.path.isfile(desertation):
                            os.remove(desertation)
                download_try_count += 1
                download_flag = all(download_flag_store)
            if not download_flag:
                raise ValueError("Cannot download '{:s}' data.".format(subject.ID))
            
    def download_support_files(self, total_retry_time = 10):
        """
        Download all support files
        """
        if self.support_files is not None and self.path_support_file is not None:
            for file_name in self.support_files:
                download_try_count = 0
                download_flag = False
                while (not download_flag) and (download_try_count < total_retry_time):
                    if download_try_count>0:
                        warnings.warn("There is an error when donwloading '{:s}'. So retry ({:n}) after 10 seconds.".format(file_name, download_try_count))
                        time.sleep(10)
                    download_flag, source_url, desertation = self.download_file(file_name)
                    if (not download_flag):
                        if os.path.isfile(desertation):
                            os.remove(desertation)
                    download_try_count += 1
                if not download_flag:
                    raise ValueError("Cannot download '{:s}'.".format(file_name))
    
    def leave_one_block_out(self,
                            block_idx: int) -> Tuple[List[int], List[int]]:
        """
        Generate testing and training blocks for specific block based on leave-one-out rule

        Parameters
        ----------
        block_idx : int
            Specific block index

        Returns
        -------
        test_block: List[int]
            Testing block
        train_block: List[int]
            Training block
        """
        if block_idx < 0:
            raise ValueError('Block index cannot be negative')
        if block_idx > self.block_num-1:
            raise ValueError('Block index should be smaller than {:d}'.format(self.block_num-1))
            
        test_block = [block_idx]
        train_block = [i for i in range(self.block_num)]
        train_block.remove(block_idx)
        
        return test_block, train_block

    def get_ch_idx(self,
                   ch_name : str):
        for ch_idx, ch_val in enumerate(self.channels):
            if ch_val.upper() == ch_name.upper():
                return ch_idx
        return None
    
    def get_data(self,
                 sub_idx: int,
                 blocks: List[int],
                 trials: List[int],
                 channels: List[int],
                 sig_len: float,
                 t_latency: Optional[float] = None,
                 shuffle: bool = False) -> Tuple[List[ndarray], List[int]]:
        """
        Construct data, corresponding labels, and sine-cosine-based reference signals 
        from one subject (specific stimuli)

        Parameters
        ----------
        sub_idx : int
            Subject index
        blocks:
            List of block indx
            Note - The index of the 1st block is 0
        trials : List[int]
            List of trial indx
            Note - The index of the 1st trial is 0
        channels: List[int]
            List of channels
        sig_len : float
            signal length (in second)
        t_latency : Optional[float]
            latency time (in second)
        shuffle : Optional[bool]
            Whether shuffle trials

        Returns
        -------
        X: List[ndarray]
            List of single trial data
        Y: List[int]
            List of corresponding label (stimulus idx)
            Note - The index of the 1st stimulus is 0
        """
        if type(sub_idx) is not list:
            sub_idx = [sub_idx]
        if t_latency is None:
            t_latency = self.default_t_latency
        if type(blocks) is not list:
            blocks = [blocks]
        if type(trials) is not list:
            trials = [trials]
        
        X = []
        Y = []

        for sub_idx_value in sub_idx:
            if type(sub_idx_value) is not int:
                raise ValueError("Subject indices must be integers")
            sub_data = self.get_sub_data(sub_idx_value)
            for block_idx in blocks:
                X.extend(self.get_data_trial(sub_data, block_idx, trials, channels, sig_len, t_latency))
                Y.extend(self.get_label_trial(sub_idx_value,block_idx,trials))
        
        if shuffle:
            trial_seq = [i for i in range(len(X))]
            random.shuffle(trial_seq)
            return [X[i] for i in trial_seq], [Y[i] for i in trial_seq]
        else:
            return X, Y
    
    def get_data_all_trials(self,
                            sub_idx: int,
                            blocks: List[int],
                            channels: List[int],
                            sig_len: float,
                            t_latency: Optional[float] = None,
                            shuffle: Optional[bool] = False) -> Tuple[List[ndarray], List[int]]:
        """
        Construct data, corresponding labels, and sine-cosine-based reference signals 
        from one subject (all trials)

        Parameters
        ----------
        sub_idx : int
            Subject index
        blocks:
            List of block indx
            Note - The index of the 1st block is 0
        channels: List[int]
            List of channels
        sig_len : float
            signal length (in second)
        t_latency : Optional[float]
            latency time (in second)
        shuffle : Optional[bool]
            Whether shuffle trials

        Returns
        -------
        X: List[ndarray]
            List of single trial data
        Y: List[int]
            List of corresponding label (stimulus idx)
            Note - The index of the 1st stimulus is 0
        """
        if type(blocks) is not list:
            blocks = [blocks]
        trials = [i for i in range(self.trial_num)]
            
        X, Y = self.get_data(sub_idx, blocks, trials, channels, sig_len, t_latency)
        
        trial_seq = [i for i in range(len(X))]
        if shuffle:
            random.shuffle(trial_seq)
        
        return [X[i] for i in trial_seq], [Y[i] for i in trial_seq]
    
    def get_data_single_trial(self,
                             sub_data: ndarray,
                             block_idx: int,
                             trial_idx: int,
                             channels: List[int],
                             sig_len: float,
                             t_latency: float) -> ndarray:
        """
        Get single trial data

        Parameters
        ----------
        sub_data : ndarray
            Subject data
            block_num * trial_num * ch_num * whole_trial_samples
        block_idx : int
            Block index
        trial_idx : int
            Trial index
        channels: List[int]
            List of channels
        sig_len : float
            signal length (in second)
        t_latency : float
            latency time (in second)

        Returns
        -------
        single_trial_data: ndarray
            filterbank_num * ch_num * (sig_len * self.srate)

        """
        if block_idx < 0:
            raise ValueError('Block index cannot be negative')
        if block_idx > self.block_num-1:
            raise ValueError('Block index should be smaller than {:d}'.format(self.block_num))
            
        if trial_idx < 0:
            raise ValueError('Stimulus index cannot be negative')
        if trial_idx > self.trial_num-1:
            raise ValueError('Stimulus index should be smaller than {:d}'.format(self.trial_num))
        
        if type(channels) is not list:
            channels = [channels]
        min_ch_idx = min(channels)
        if min_ch_idx < 0:
            raise ValueError('Channel index cannot be negative')
        max_ch_idx = max(channels)
        if max_ch_idx > len(self.channels)-1:
            raise ValueError('Channel index should be smaller than {:d}'.format(len(self.channels)))
            
        if sig_len < 0 or t_latency < 0:
            raise ValueError('Time cannot be negative')
        if sig_len + t_latency + self.t_prestim > self.trial_len:
            # raise ValueError('Total time length cannot be larger than single trial time')
            warnings.warn('Total time length ({:n}) is larger than single trial time ({:n})'.format(sig_len + t_latency, self.trial_len))
            
        t_pre = floor((self.t_prestim + t_latency) * self.srate)
        t_prestim = floor(self.t_prestim * self.srate)
        t_latency = floor(t_latency * self.srate)
        sig_len = floor(sig_len * self.srate)
        
        start_t_idx = t_prestim
        end_t_idx = t_pre + sig_len
        eeg_total = sub_data[block_idx,trial_idx,channels,start_t_idx:end_t_idx]
        # end_t_idx = eeg_total.shape[-1]
        eeg_total = self.preprocess_fun(self, eeg_total)
        eeg_total = self.filterbank_fun(self, eeg_total)

        if len(eeg_total.shape) == 1:
            eeg_total = expand_dims(eeg_total, axis = 0)
            eeg_total = expand_dims(eeg_total, axis = 0)
        elif len(eeg_total.shape) == 2:
            eeg_total = expand_dims(eeg_total, axis = 0)
        elif len(eeg_total.shape) == 3:
            pass
        else:
            raise ValueError("EEG shape has error. Please check the output of the filterbank function should have three dimention. The 1st dimention is the filterbank number")
        
        start_t_idx = t_latency
        end_t_idx = t_latency + sig_len
        eeg_trial = eeg_total[:,:,start_t_idx:end_t_idx]
        
        return eeg_trial
    
    def get_ref_sig(self,
                    sig_len: float,
                    N: int,
                    ignore_stim_phase: bool = False) -> List[ndarray]:
        """
        Construct sine-cosine-based reference signals  for all stimuli
        
        Parameters
        ----------
        sig_len : float
            signal length (in second)
        N : int
            Number of harmonics

        Returns
        -------
        ref_sig : List[ndarray]
            List of reference signals
        """
        if ignore_stim_phase:
            phases = [0 for _ in range(len(self.stim_info['freqs']))]
        else:
            phases = self.stim_info['phases']
        ref_sig = self.ref_sig_fun(self, sig_len, N, phases)
        return ref_sig
    
    def regist_preprocess(self,
                          preprocess_fun):
        """
        Regist preprocess function

        Parameters
        ----------
        preprocess_fun
            preprocess function
            
            Parameters
            ----------
            dataself
                data instance
            X : ndarray
                EEG signal (including latency time and desired signal window, but pre-stimulus time has been removed)
                ch_num * signal_length
                
            Returns
            ----------
            preprocess_X: ndarray
        """
        self.preprocess_fun = preprocess_fun

    def regist_ref_sig_fun(self,
                           ref_sig_fun):
        """
        Regist reference signal generation function

        Parameters
        ------------
        ref_sig_fun
            reference signal generation function

            Parameters
            -----------
            dataself
                data instance
            sig_len : float
                signal length (in second)
            N : int
                Number of harmonics
            phases : List[float]
                Phases of references

            Returns
            -------
            ref_sig : List[ndarray]
                List of reference signals
        """
        self.ref_sig_fun = ref_sig_fun
    
    def regist_filterbank(self,
                          filterbank_fun):
        """
        Regist filterbank function

        Parameters
        ----------
        filterbank_fun
            filterbank function
            
            Parameters
            ----------
            dataself
                data instance
            X : ndarray
                EEG signal after preprocess (including latency time and desired signal window, but pre-stimulus time has been removed)
                ch_num * signal_length

            Returns
            -------
            filterbank_X: ndarray
                filterbanks
        """
        self.filterbank_fun = filterbank_fun

    def get_data_trial(self,
                       sub_data : ndarray, 
                       block_idx : int, 
                       trials : List[int], 
                       channels : List[int], 
                       sig_len : float, 
                       t_latency : float):
        return [self.get_data_single_trial(sub_data, block_idx, trial_idx, channels, sig_len, t_latency) for trial_idx in trials]

    def get_label_trial(self,
                        sub_idx : int,
                        block_idx : int,
                        trials : List[int]):
        """
        Get labels of given trials
        """
        return [self.get_label_single_trial(sub_idx, block_idx, trial_idx) for trial_idx in trials]

    def get_snr_single_trial(self,
                             sub_idx : int,
                             block_idx : int,
                             trial_idx : int,
                             ch_idx : int,
                             sig_len : float,
                             filter_bank_idx : int = 0,
                             Nh : int = 1,
                             srate : Optional[float] = None,
                             t_latency: Optional[float] = None,
                             detrend_flag : bool = True,
                             NFFT : Optional[int] = None,
                             type : str = 'fft',
                             harmonic_num : int = 5,
                             ignore_stim_phase : bool = False):
        """
        Calculate the SNR of one single trial
        """
        if srate is None:
            srate = self.srate
        X_fft, Y_fft = self.get_data(sub_idx = sub_idx,
                                        blocks = [block_idx],
                                        trials = [trial_idx],
                                        channels = [ch_idx],
                                        sig_len = sig_len,
                                        t_latency = t_latency)
        X_fft = X_fft[0][filter_bank_idx,:,:]
        Y_fft = Y_fft[0]
        if type.lower() == 'fft':
            return freqs_snr(X_fft, self.stim_info['freqs'][Y_fft], srate, Nh, 
                            detrend_flag = detrend_flag,
                            NFFT = NFFT)
        elif type.lower() == 'sine':
            ref_sig = self.get_ref_sig(sig_len, harmonic_num, ignore_stim_phase)
            return sine_snr(X_fft, ref_sig[Y_fft])
        else:
            raise ValueError('Unknown SNR type')

    def get_phase_single_trial(self,
                             sub_idx : int,
                             block_idx : int,
                             trial_idx : int,
                             ch_idx : int,
                             sig_len : float,
                             filter_bank_idx : int = 0,
                             srate : Optional[float] = None,
                             t_latency: Optional[float] = None,
                             detrend_flag : bool = True,
                             NFFT : Optional[int] = None,
                             remove_target_phase : bool = False):
        """
        Calculate the phase of one single trial
        """
        if srate is None:
            srate = self.srate
        X_fft, Y_fft = self.get_data(sub_idx = sub_idx,
                                        blocks = [block_idx],
                                        trials = [trial_idx],
                                        channels = [ch_idx],
                                        sig_len = sig_len,
                                        t_latency = t_latency)
        X_fft = X_fft[0][filter_bank_idx,:,:]
        Y_fft = Y_fft[0]
        phase = freqs_phase(X_fft, self.stim_info['freqs'][Y_fft], self.stim_info['phases'][Y_fft],
                           srate, 
                           detrend_flag = detrend_flag,
                           NFFT = NFFT)
        if remove_target_phase:
            phase -= self.stim_info['phases'][Y_fft]
        return phase

    def get_snr(self,
                filter_bank_idx : int = 0,
                Nh : int = 1,
                srate : Optional[float] = None,
                t_latency: Optional[float] = None,
                remove_break : bool = True,
                remove_pre_and_latency : bool = True,
                display_progress : bool = False,
                detrend_flag : bool = True,
                NFFT : Optional[int] = None,
                sig_len : Optional[float] = None,
                type : str = 'fft',
                harmonic_num : int = 5,
                ignore_stim_phase : bool = False,
                ch_used_recog : List[int] = None):
        """
        Calculate the SNR
        """
        if srate is None:
            srate = self.srate
        if t_latency is None:
            t_latency = self.default_t_latency
        if sig_len is None:
            sig_len = self.trial_len
        
        if remove_pre_and_latency:
            sig_len = sig_len - self.t_prestim - t_latency
        if remove_break:
            sig_len -= self.t_break
        if display_progress:
            pbar = create_pbar([len(self.subjects), self.block_num])
        
        if type.lower() == 'fft':
            snr = np.zeros((len(self.subjects), self.block_num, self.trial_num, len(self.channels))) # subj * block_num * stimulus_num * ch_num
        elif type.lower() == 'sine':
            if ch_used_recog is None:
                raise ValueError("'ch_used_recog' cannot be None.")
            snr = np.zeros((len(self.subjects), self.block_num, self.trial_num)) # subj * block_num * stimulus_num
            ref_sig = self.get_ref_sig(sig_len, harmonic_num, ignore_stim_phase)
        else:
            raise ValueError('Unknown SNR type')

        for sub_idx in range(len(self.subjects)):
            for block_idx in range(self.block_num):
                if display_progress:
                    pbar.update(1)
                X_all_trials, Y_all_trials = self.get_data_all_trials(sub_idx, [block_idx], list(range(len(self.channels))),sig_len, t_latency)
                for trial_idx, X_single_trial in enumerate(X_all_trials):
                    Y_fft = Y_all_trials[trial_idx]
                    if type.lower() == 'fft':
                        for ch_idx in range(len(self.channels)):
                            X_fft = X_single_trial[filter_bank_idx, ch_idx:(ch_idx+1), :]
                            snr[sub_idx, block_idx, trial_idx, ch_idx] = freqs_snr(X_fft, self.stim_info['freqs'][Y_fft], srate, Nh,
                                                                                    detrend_flag = detrend_flag,
                                                                                    NFFT = NFFT)
                    elif type.lower() == 'sine':
                        X_fft = X_single_trial[filter_bank_idx, ch_used_recog, :]
                        snr[sub_idx, block_idx, trial_idx] = sine_snr(X_fft, ref_sig[Y_fft])
                    else:
                        raise ValueError('Unknown SNR type')
        return snr

    def get_phase(self,
                filter_bank_idx : int = 0,
                srate : Optional[float] = None,
                t_latency: Optional[float] = None,
                remove_break : bool = True,
                remove_pre_and_latency : bool = True,
                display_progress : bool = False,
                detrend_flag : bool = True,
                NFFT : Optional[int] = None,
                sig_len : Optional[float] = None,
                remove_target_phase : bool = False):
        """
        Calculate the SNR
        """
        if srate is None:
            srate = self.srate
        if t_latency is None:
            t_latency = self.default_t_latency
        if sig_len is None:
            sig_len = self.trial_len
        phase = np.zeros((len(self.subjects), self.block_num, self.trial_num, len(self.channels))) # subj * block_num * stimulus_num * ch_num
        if remove_pre_and_latency:
            sig_len = sig_len - self.t_prestim - t_latency
        if remove_break:
            sig_len -= self.t_break
        if display_progress:
            pbar = create_pbar([len(self.subjects), self.block_num])
        for sub_idx in range(len(self.subjects)):
            for block_idx in range(self.block_num):
                if display_progress:
                    pbar.update(1)
                X_all_trials, Y_all_trials = self.get_data_all_trials(sub_idx, [block_idx], list(range(len(self.channels))),sig_len, t_latency)
                for trial_idx, X_single_trial in enumerate(X_all_trials):
                    Y_fft = Y_all_trials[trial_idx]
                    for ch_idx in range(len(self.channels)):
                        X_fft = X_single_trial[filter_bank_idx, ch_idx:(ch_idx+1), :]
                        phase[sub_idx, block_idx, trial_idx, ch_idx] = freqs_phase(X_fft, self.stim_info['freqs'][Y_fft], self.stim_info['phases'][Y_fft],
                                                                                srate, 
                                                                                detrend_flag = detrend_flag,
                                                                                NFFT = NFFT)
                        if remove_target_phase:
                            phase[sub_idx, block_idx, trial_idx, ch_idx] -= self.stim_info['phases'][Y_fft]
        return phase
            
    def __repr__(self):
        return self.__str__()
    
    def __str__(self):
        desc = """Dataset: {:s}:\n   Subjects: {:d}\n   Srate: {:d}\n   Channels: {:d}\n   Blocks: {:d}\n  Trials: {:d}\n   Stimuli: {:d}\n   Signal length: {:.3f}\n""".format(
            self.ID,
            len(self.subjects),
            self.srate,
            len(self.channels),
            self.block_num,
            self.trial_num,
            self.stim_info['stim_num'],
            self.trial_len
            )
        return desc
    
    @abc.abstractclassmethod
    def download_single_subject(self,
                                subject: SubInfo):
        """
        Download single subject's data
        
        parameters
        --------------
        subject: SubInfo
            Subject information

        """
        pass
    
    @abc.abstractclassmethod
    def download_file(self,
                      file_name: str):
        """
        Download specific file

        Parameters
        ----------
        file_name : str
            File name
        """
        pass
    
    
    @abc.abstractclassmethod
    def get_sub_data(self, 
                     sub_idx: int) -> ndarray:
        """
        Get single subject data

        Parameters
        ----------
        sub_idx : int
            Subject index

        Returns
        -------
        sub_data : ndarray
            Subject data
            block_num * stimulus_num * ch_num * whole_trial_samples
        """
        pass
    
    @abc.abstractclassmethod
    def get_label_single_trial(self,
                               sub_idx: int,
                               block_idx: int,
                               trial_idx: int) -> int:
        """
        Get the label of single trial

        Parameters
        ----------
        sub_idx : int
            Subject index
        block_idx : int
            Block index
        trial_idx : int
            Trial index

        Returns
        -------
        label: int
            Label of stimulus
        """
        pass
    
    
    


# -*- coding: utf-8 -*-

import os
from typing import Union, Optional, Dict, List, Tuple

from pooch import retrieve

def download_single_file(source_url: str, 
                         desertation: str,
                         known_hash: Optional[str] = None,
                         progressbar: bool = True):
    """
    Download one file

    Parameters
    ----------
    source_url : str
        Source url
    desertation : str
        Local file path
    """
    
    file_name = os.path.basename(desertation)
    desertation_dir = os.path.dirname(desertation)
    
    if not os.path.exists(desertation_dir):
        os.makedirs(desertation_dir)
    if os.path.isfile(desertation):
        os.remove(desertation)
        
    retrieve(source_url, known_hash, 
             fname = file_name,
             path = desertation_dir,
             progressbar=progressbar)

    if not os.path.exists(desertation):
        raise ValueError("The following file has been downloaded but cannot find it: {:s}".format(desertation))
    


# -*- coding: utf-8 -*-

from typing import Union, Optional, Dict, List, Tuple, Callable
from numpy import iscomplex, ndarray

import numpy as np
import numpy.linalg as nplin
import scipy.linalg as slin

from sklearn import linear_model as sklinmodel
from warnings import simplefilter
from sklearn.exceptions import ConvergenceWarning
# simplefilter(action='ignore', category=ConvergenceWarning)
# @ignore_warnings(category=ConvergenceWarning)

import warnings 

from .utils import (
    eigvec, svd, get_depend_column_ind
)

from SSVEPAnalysisToolbox.utils.io import savedata

# import tensorflow as tf
# tf.compat.v1.disable_eager_execution()
# gpu = tf.config.list_physical_devices('GPU')
# tf.config.experimental.set_memory_growth(gpu[0], True)

def combine_name(method_self):
    return method_self.ID[0:-1] + '-' + method_self.LSconfig['LSMethod'] + ')'

def ssvep_lsframe(Z : ndarray,
                  Lx : ndarray,
                  Px : Optional[ndarray] = None,
                  Ly : Optional[ndarray] = None,
                  Py : Optional[ndarray] = None,
                  LSMethod: str = 'lstsq',
                  displayConvergWarn: bool = False,
                  max_n: int = 100,
                  M_error_threshold: float = 1e-6,
                  W_error_threshold: float = 1e-6,
                  smallest_n: int = 1,
                  largest_n: int = 100,
                  check_rank: bool = True,
                  l1_alpha = None,
                  l2_alpha = None,
                  alpha = None,
                  reg_iter = 1000,
                  reg_tol = 1e-4,
                  kernel_fun = None):
    if Px is None:
        _, Dx, Vx = svd(Lx, False, True)
        PKZ = Z.copy()
    else:
        if Ly is None and Py is None:
            raise ValueError("When Px is not None, Ly and Py are also should not be None.")
        try:
            LP = Ly @ Py
        except:
            LP = Ly * Py
        try:
            PKZ = LP @ Z
        except:
            PKZ = LP * Z
        _, Dx, Vx = svd(PKZ, False, True)
        PKZ = Lx @ Px @ Z
    X_rank = nplin.matrix_rank(PKZ)
    if check_rank:
        if X_rank!=min(PKZ.shape):
            max_n = smallest_n
        else:
            max_n = largest_n
    else:
        max_n = largest_n
    # max_n = largest_n
    small_value = 1e-50
    for zero_i in np.where(Dx==0)[0]:
        if zero_i == 0:
            Dx[zero_i] = small_value
        else:
            while True:
                if Dx[zero_i-1] > small_value:
                    Dx[zero_i] = small_value
                    break
                else:
                    small_value = small_value/10
    Dx_inv = nplin.inv(np.diag(Dx))
    VDV = Vx.T @ Dx_inv @ Vx
    W, M, M_error, W_error, n = lsframe(PKZ, 
                                        PKZ @ VDV, 
                                        LSMethod = LSMethod,
                                        max_n = max_n,
                                        displayConvergWarn = displayConvergWarn,
                                        M_error_threshold = M_error_threshold,
                                        W_error_threshold = W_error_threshold,
                                        l1_alpha = l1_alpha,
                                        l2_alpha = l2_alpha,
                                        alpha = alpha,
                                        reg_iter = reg_iter,
                                        reg_tol = reg_tol,
                                        kernel_fun = kernel_fun,
                                        check_full_rank = check_rank)
    if check_rank:
        if X_rank!=min(PKZ.shape):
            return W, M, M_error, W_error, n, False
        else:
            return W, M, M_error, W_error, n, True
        # W, M, M_error, W_error, n, nplin.matrix_rank(PKZ)==min(PKZ.shape)
    else:
        return W, M, M_error, W_error, n, None

def sigmoid(x):
    return 1 / (1 + np.exp(-x))

def tanh(x):
    s1 = np.exp(x) - np.exp(-x)
    s2 = np.exp(x) + np.exp(-x)
    s = s1 /s2
    return s

def relu(x, thr = 0, lowest_bound = 0):
    s = np.where(x < thr, lowest_bound, x)
    return s

def binary(x, thr = 0, lowest_bound = 0):
    s = np.where(x < thr, lowest_bound, 1)
    return s

def softmax(x):
    return np.exp(x) / np.sum(np.exp(x), axis = 0)

def lsframe(X: ndarray,
            Y: ndarray,
            LSMethod: str = 'lstsq',
            displayConvergWarn: bool = False,
            max_n: int = 100,
            M_error_threshold: float = 1e-6,
            W_error_threshold: float = 1e-6,
            l1_alpha = None,
            l2_alpha = None,
            alpha = None,
            reg_iter = 1000,
            reg_tol = 1e-4,
            kernel_fun = None,
            check_full_rank = False) -> Tuple[list,list,list,list,int]:
    """
    Solve XWM'=Y

    Parameters
    ----------
    X : ndarray
    Y : ndarray

    Returns
    -------
    W : list
        W values of all iterations
    M : list
        M values of all iterations
    M_error : list
        M_error of all iterations
    W_error : list
        W_error of all iterations
    n : int
        Total iteration number
    """
    if alpha is None:
        alpha = 1
    if kernel_fun is not None:
        # Normalization
        X_norm = np.linalg.norm(X, axis = 1, keepdims=True)
        X = X / X_norm
        Y_norm = np.linalg.norm(Y, axis = 1, keepdims=True)
        Y = Y / Y_norm
        # Apply kernel
        if kernel_fun.lower() == 'sigmoid':
            X = sigmoid(X)
            Y = sigmoid(Y)
        elif kernel_fun.lower() == 'tanh':
            X = tanh(X)
            Y = tanh(Y)
        elif kernel_fun.lower() == 'relu':
            X = relu(X)
            Y = relu(Y)
        elif kernel_fun.lower() == 'binary':
            X = binary(X)
            Y = binary(Y)
        else:
            raise ValueError('Unknown kernel type')
    W = []
    M = []
    n = 0
    M_error = []
    W_error = []
    # Initialize M
    # M_init = eigvec(Y.T @ X @ nplin.inv(X.T @ X) @ X.T @ Y)
    M_init = eigvec(Y.T @ Y)
    # M_tmp = np.flip(M_tmp,1)
    M.append(M_init)
    W.append(None)
    M_error.append(np.inf)
    W_error.append(np.inf)
    while n < max_n and (M_error[-1] > M_error_threshold and W_error[-1] > W_error_threshold):
        n += 1
        # fix M and update W
        if LSMethod.lower() == 'lstsq':
            # if check_full_rank:
            #     remove_column_X = get_depend_column_ind(X)
            #     keep_column_X = np.delete(list(range(X.shape[1])), remove_column_X)
            #     YM = Y @ M[n-1]
            #     W_tmp = np.zeros((X.shape[1],YM.shape[1]))
            #     W_tmp[keep_column_X,:], _, _, _ = nplin.lstsq(X[:,keep_column_X], YM, rcond=None)
            #     if len(remove_column_X)!=0:
            #         W_tmp[remove_column_X,:], _, _, _ = nplin.lstsq(X[:,remove_column_X], YM, rcond=None)
            # else:
            #     W_tmp, _, _, _ = nplin.lstsq(X, Y @ M[n-1], rcond=None)
            try:
                W_tmp, _, _, _ = nplin.lstsq(X, Y @ M[n-1], rcond=None)
            except nplin.LinAlgError:
                # X_new = X[:,::-1]
                remove_column_X = get_depend_column_ind(X)
                keep_column_X = np.delete(list(range(X.shape[1])), remove_column_X)
                YM = Y @ M[n-1]
                W_tmp = np.zeros((X.shape[1],YM.shape[1]))
                W_tmp[keep_column_X,:], _, _, _ = nplin.lstsq(X[:,keep_column_X], YM, rcond=None)
                # W_tmp = W_tmp[::-1,:]
                # data = {"X": X,
                # "Y": Y,
                # "M": M,
                # "remove_column_X": remove_column_X,
                # "keep_column_X": keep_column_X,
                # "W_tmp": W_tmp}
                # data_file = 'test_temp.mat'
                # savedata(data_file, data, 'mat')
                # raise ValueError('Store test data when error')
        elif LSMethod.lower() == 'ols':
            # if check_full_rank:
            #     remove_column_X = get_depend_column_ind(X)
            #     keep_column_X = np.delete(list(range(X.shape[1])), remove_column_X)
            #     YM = Y @ M[n-1]
            #     W_tmp = np.zeros((X.shape[1],YM.shape[1]))
            #     W_tmp[keep_column_X,:] = ols(X[:,keep_column_X], YM)
            # else:
            #     W_tmp = ols(X, Y @ M[n-1])
            W_tmp = ols(X, Y @ M[n-1])
        elif LSMethod.lower() == 'lasso':
            if l1_alpha is None:
                raise ValueError("'l1_alpha' must be given for lasso.")
            W_tmp = lasso_sklearn(X, Y @ M[n-1], alpha * l1_alpha, 
                                  max_iter = reg_iter,
                                  tol = reg_tol)
            W_tmp = W_tmp.T
        # elif LSMethod.lower() == 'lasso-gpu':
        #     if l1_alpha is None:
        #         raise ValueError("'l1_alpha' must be given for lasso.")
        #     W_tmp, _ = ls_ElasticNet_gpu(X, Y @ M[n-1], 1,
        #                                 alpha = alpha * l1_alpha, 
        #                                 fit_intercept = False,
        #                                 max_iter = reg_iter,
        #                                 tol = reg_tol)
        elif LSMethod.lower() == 'elastic_net':
            if l1_alpha is None:
                raise ValueError("'l1_alpha' must be given for elastic_net.")
            W_tmp = elastic_net_sklearn(X, Y @ M[n-1], l1_alpha, 
                                        alpha = alpha,
                                        max_iter = reg_iter,
                                        tol = reg_tol)
            W_tmp = W_tmp.T
        elif LSMethod.lower() == 'ridge':
            if l2_alpha is None:
                raise ValueError("'l2_alpha' must be given for ridge.")
            W_tmp = ridge_sklearn(X, Y @ M[n-1], alpha * l2_alpha, 
                                  max_iter = reg_iter,
                                  tol = reg_tol)
            W_tmp = W_tmp.T
        else:
            raise ValueError("Unknown LSMethod!!")
        D_tmp = W_tmp.T @ X.T @ X @ W_tmp
        D_tmp = np.diagonal(D_tmp)
        sort_i = np.argsort(D_tmp)
        sort_i_flip = np.flip(sort_i)
        W_tmp = W_tmp[:,sort_i_flip]
        # fix W and update M
        M_tmp = np.zeros_like(M_init)
        for i in range(M_tmp.shape[1]):
            Z = Y.T @ X @ W_tmp[:,i:(i+1)]
            Uz, Dz, Vz = svd(Z, False, True)
            M_tmp[:,i:(i+1)] = Uz @ Vz
        W.append(W_tmp.copy())
        M.append(M_tmp.copy())
        # check convergence
        M_error.append(
            np.sum(np.abs(M[n]-M[n-1]))
        )
        if n==1:
            W_error.append(M_error[-1])
        else:
            W_error.append(
                np.sum(np.abs(W[n]-W[n-1]))
            )
    if n>=max_n and displayConvergWarn:
        warnings.warn("Cannot converge")
    return W, M, M_error, W_error, n

def get_lsconfig(LSconfig):
    kernel_fun = None
    if LSconfig is None:
        LSMethod = 'lstsq'
        displayConvergWarn = False
        max_n = 100
        M_error_threshold = 1e-6
        W_error_threshold = 1e-6
        l1_alpha = None
        l2_alpha = None
        alpha = None
        reg_iter = 1000
        reg_tol = 1e-4
    else:
        if len(LSconfig.values()) == 10:
            LSMethod, displayConvergWarn, max_n, M_error_threshold, W_error_threshold, l1_alpha, l2_alpha, alpha, reg_iter, reg_tol = LSconfig.values()
        elif len(LSconfig.values()) == 11:
            LSMethod, displayConvergWarn, max_n, M_error_threshold, W_error_threshold, l1_alpha, l2_alpha, alpha, reg_iter, reg_tol, kernel_fun = LSconfig.values()
        else:
            raise ValueError('The number of configures in LSconfig is wrong.')
    if LSMethod.lower() == 'lstsq' or LSMethod.lower() == 'ols':
        check_full_rank = True
    else:
        check_full_rank = False
    return LSMethod, displayConvergWarn, max_n, M_error_threshold, W_error_threshold, l1_alpha, l2_alpha, alpha, reg_iter, reg_tol, check_full_rank, kernel_fun

def canoncorr_ls(X: ndarray,
                 Y: ndarray,
                 force_output_UV: Optional[bool] = False,
                 LSconfig: Optional[dict] = None) -> Union[Tuple[ndarray, ndarray, ndarray], ndarray]:
    """
    Canonical correlation analysis based on LS framework
    """
    LSMethod, displayConvergWarn, max_n, M_error_threshold, W_error_threshold, l1_alpha, l2_alpha, alpha, reg_iter, reg_tol, check_full_rank, kernel_fun = get_lsconfig(LSconfig)
    X = X - np.mean(X,0)
    Y = Y - np.mean(Y,0)
    # Parameters of unified framework
    Z = X.copy()
    K = np.eye(X.shape[0])
    Q, _, _ = slin.qr(Y, mode = 'economic', pivoting = True)
    P = Q @ Q.T
    # LS framework for spatial filter
    _, Dx, Vx = svd(K.T @ Z, False, True)
    Dx = np.diag(Dx)
    PKZ = P.T @ K.T @ Z
    X_rank = nplin.matrix_rank(PKZ)
    if X_rank!=min(PKZ.shape) and check_full_rank:
        applied_max_n = 1
    else:
        applied_max_n = max_n
    W, _, _, _, _ = lsframe(PKZ, 
                            PKZ @ Vx.T @ nplin.inv(Dx) @ Vx, 
                            max_n = applied_max_n,
                            LSMethod = LSMethod,
                            displayConvergWarn = displayConvergWarn,
                            M_error_threshold = M_error_threshold,
                            W_error_threshold = W_error_threshold,
                            l1_alpha = l1_alpha,
                            l2_alpha = l2_alpha,
                            alpha = alpha,
                            reg_iter = reg_iter,
                            reg_tol = reg_tol,
                            kernel_fun = kernel_fun,
                            check_full_rank = check_full_rank)
    W = W[-1].copy()
    # Weights of reference signal and correlation coefficients
    V, _, _, _ = nplin.lstsq(Y, X @ W, rcond=None)
    if X_rank!=min(PKZ.shape) and check_full_rank:
        W, _, _, _ = nplin.lstsq(Z, Y @ V, rcond=None)
    r = []
    for i in range(W.shape[1]):
        r.append(
            np.corrcoef(W[:,i:(i+1)].T @ X.T, V[:,i:(i+1)].T @ Y.T)[0,1]
        )
    r = np.array(r)
    # return results
    if force_output_UV:
        return W, V, r
    else:
        return r

def ols(X, y, fit_intercept=False):
    """Ordinary Least Squares (OLS) Regression model with intercept term.
    Fits an OLS regression model using the closed-form OLS estimator equation.
    Intercept term is included via design matrix augmentation.
    Params:
        X - NumPy matrix, size (N, p), of numerical predictors
        y - NumPy array, length N, of numerical response
        fit_intercept - Boolean indicating whether to include an intercept term
    Returns:
        NumPy array, length p + 1, of fitted model coefficients
    """
    if fit_intercept:
        X = np.hstack((np.ones((X.shape[0], 1)), X))
    _, n_y = y.shape
    w = []
    for i in range(n_y):
        w.append(nplin.solve(np.dot(X.T, X), np.dot(X.T, y[:,i])))
    w = [np.expand_dims(tmp,axis = 1) for tmp in w]
    w = np.concatenate(w, axis = 1)
    return w

def lasso_sklearn(X, y, l1, 
                  fit_intercept = False,
                  max_iter = 1000,
                  tol = 1e-4):
    clf = sklinmodel.Lasso(alpha = l1,
                           fit_intercept = fit_intercept,
                           max_iter = max_iter,
                           tol = tol,
                           precompute = True)
    # alphas, clf, dual_gaps = sklinmodel.lasso_path(X, y, eps = 1e-5)
    # clf = clf[0,:,:]
    simplefilter(action='ignore', category=ConvergenceWarning)
    clf.fit(X, y)
    if fit_intercept:
        return np.concatenate(clf.coef_, np.expand_dims(clf.intercept_, axis = 1), axis = 0)
    return clf.coef_

def lasso(X, y, l1, tol=1e-6, path_length=100, return_path=False, fit_intercept=False):
    """The Lasso Regression model with intercept term.
    Intercept term included via design matrix augmentation.
    Pathwise coordinate descent with co-variance updates is applied.
    Path from max value of the L1 tuning parameter to input tuning parameter value.
    Features must be standardized (centered and scaled to unit variance)
    Params:
        X - NumPy matrix, size (N, p), of standardized numerical predictors
        y - NumPy array, length N, of numerical response
        l1 - L1 penalty tuning parameter (positive scalar)
        tol - Coordinate Descent convergence tolerance (exited if change < tol)
        path_length - Number of tuning parameter values to include in path (positive integer)
        return_path - Boolean indicating whether model coefficients along path should be returned
    Returns:
        if return_path == False:
            NumPy array, length p + 1, of fitted model coefficients
        if return_path == True:
            List, length 3, of last fitted model coefficients, tuning parameter path and coefficient values
    Reference:
        https://towardsdatascience.com/regularized-linear-regression-models-dcf5aa662ab9
    """
    if fit_intercept:
        X = np.hstack((np.ones((len(X), 1)), X))
    m, n = np.shape(X)
    B_star = np.zeros((n))
    l_max = max(list(abs(np.dot(np.transpose(X[:, 1:]), y)))) / m
    # At or above l_max, all coefficients (except intercept) will be brought to 0
    if l1 >= l_max:
        if fit_intercept:
            B_star = np.append(np.mean(y), np.zeros((n - 1)))
        else:
            B_star = np.zeros(n)
        if return_path:
            return [B_star, None, None]
        else:
            return B_star
    l_path = np.geomspace(l_max, l1, path_length)
    coeffiecients = np.zeros((len(l_path), n))
    for i in range(len(l_path)):
        while True:
            B_s = B_star
            for j in range(n):
                k = np.where(B_s != 0)[0]
                update = (1/m)*((np.dot(X[:,j], y)- \
                                np.dot(np.dot(X[:,j], X[:,k]), B_s[k]))) + \
                                B_s[j]
                B_star[j] = (np.sign(update) * max(abs(update) - l_path[i], 0))
            if np.all(abs(B_s - B_star) < tol):
                coeffiecients[i, :] = B_star
                break
    if return_path:
        return [B_star, l_path, coeffiecients]
    else:
        return B_star

def elastic_net_sklearn(X, y, l1_ratio,
                        alpha = 1, 
                        fit_intercept = False,
                        max_iter = 1000,
                        tol = 1e-4):
    clf = sklinmodel.ElasticNet(alpha = alpha,
                                l1_ratio = l1_ratio,
                                fit_intercept = fit_intercept,
                                max_iter = max_iter,
                                tol = tol,
                                precompute = True)
    simplefilter(action='ignore', category=ConvergenceWarning)
    clf.fit(X, y)
    if fit_intercept:
        return np.concatenate(clf.coef_, np.expand_dims(clf.intercept_, axis = 1), axis = 0)
    return clf.coef_

def elastic_net(X, y, l, alpha, tol=1e-4, path_length=100, return_path=False, fit_intercept=False):
    """The Elastic Net Regression model with intercept term.
    Intercept term included via design matrix augmentation.
    Pathwise coordinate descent with co-variance updates is applied.
    Path from max value of the L1 tuning parameter to input tuning parameter value.
    Features must be standardized (centered and scaled to unit variance)
    Params:
        X - NumPy matrix, size (N, p), of standardized numerical predictors
        y - NumPy array, length N, of numerical response
        l - l penalty tuning parameter (positive scalar)
        alpha - alpha penalty tuning parameter (positive scalar between 0 and 1)
        tol - Coordinate Descent convergence tolerance (exited if change < tol)
        path_length - Number of tuning parameter values to include in path (positive integer)
    Returns:
        NumPy array, length p + 1, of fitted model coefficients
    Reference:
        https://towardsdatascience.com/regularized-linear-regression-models-dcf5aa662ab9
    """
    if fit_intercept:
        X = np.hstack((np.ones((len(X), 1)), X))
    m, n = np.shape(X)
    B_star = np.zeros((n))
    if alpha == 0:
        l2 = 1e-15
    l_max = max(list(abs(np.dot(np.transpose(X), y)))) / m / alpha
    if l >= l_max:
        if fit_intercept:
            B_star = np.append(np.mean(y), np.zeros((n - 1)))
        else:
            B_star = np.zeros(n)
        if return_path:
            return [B_star, None, None]
        else:
            return B_star
    l_path = np.geomspace(l_max, l, path_length)
    coeffiecients = np.zeros((len(l_path), n))
    for i in range(path_length):
        while True:
            B_s = B_star
            for j in range(n):
                k = np.where(B_s != 0)[0]
                update = (1/m)*((np.dot(X[:,j], y)- \
                                np.dot(np.dot(X[:,j], X[:,k]), B_s[k]))) + \
                                B_s[j]
                B_star[j] = (np.sign(update) * max(
                    abs(update) - l_path[i] * alpha, 0)) / (1 + (l_path[i] * (1 - alpha)))
            if np.all(abs(B_s - B_star) < tol):
                coeffiecients[i, :] = B_star
                break
    if return_path:
        return [B_star, l_path, coeffiecients]
    else:
        return B_star

def ridge_sklearn(X, y, l2, 
                 fit_intercept = False,
                 max_iter = 1000,
                 tol = 1e-4):
    clf = sklinmodel.Ridge(alpha = l2,
                           fit_intercept = fit_intercept,
                           max_iter = max_iter,
                           tol = tol)
    simplefilter(action='ignore', category=ConvergenceWarning)
    clf.fit(X, y)
    if fit_intercept:
        return np.concatenate(clf.coef_, np.expand_dims(clf.intercept_, axis = 1), axis = 0)
    return clf.coef_

def ridge(X, y, l2, fit_intercept=False):
    """Ridge Regression model with intercept term.
    L2 penalty and intercept term included via design matrix augmentation.
    This augmentation allows for the OLS estimator to be used for fitting.
    Params:
        X - NumPy matrix, size (N, p), of numerical predictors
        y - NumPy array, length N, of numerical response
        l2 - L2 penalty tuning parameter (positive scalar) 
    Returns:
        NumPy array, length p + 1, of fitted model coefficients
    Reference:
        https://towardsdatascience.com/regularized-linear-regression-models-44572e79a1b5
    """
    if fit_intercept:
        m, n = np.shape(X)
        upper_half = np.hstack((np.ones((m, 1)), X))
    else:
        upper_half = X
    lower = np.zeros((n, n))
    np.fill_diagonal(lower, np.sqrt(l2))
    lower_half = np.hstack((np.zeros((n, 1)), lower))
    X = np.vstack((upper_half, lower_half))
    y = np.append(y, np.zeros(n))
    return np.linalg.solve(np.dot(X.T, X), np.dot(X.T, y))

# def ls_ElasticNet_gpu(X, Y, l1_ratio,
#                       alpha = 1, 
#                       fit_intercept = False,
#                       max_iter = 1000,
#                       tol = 1e-4,
#                       lr = 1e-3):
#     n_samples, n_input = X.shape
#     n_epoch = max_iter
#     lam = alpha
#     n_component = Y.shape[1]
#     #
#     x = tf.compat.v1.placeholder(dtype = tf.float64, shape = [None, n_input])
#     y = tf.compat.v1.placeholder(dtype = tf.float64, shape = [None, n_component])
#     w = tf.Variable(np.random.rand(n_input, n_component))
#     b = tf.Variable(np.random.rand(1, n_component))
#     if fit_intercept:
#         z = x @ w + b
#     else:
#         z = x @ w
#     #
#     mse_loss = tf.reduce_mean((z - y) ** 2, axis = 0)
#     l1_loss = lam * l1_ratio * tf.reduce_sum(tf.abs(w), axis = 0)
#     l2_loss = lam * (1 - l1_ratio) * tf.reduce_sum(w ** 2, axis = 0)
#     if l1_ratio == 1:
#         loss = mse_loss + l1_loss
#     elif l1_ratio == 0:
#         loss = mse_loss + l2_loss
#     else:
#         loss = mse_loss + l1_loss + l2_loss
#     loss = tf.reduce_sum(loss)
#     op = tf.compat.v1.train.AdamOptimizer(lr).minimize(loss)
#     #
#     W_final = []
#     B_final = []
#     losses_final = []
#     for i in range(1):
#         losses = []
#         with tf.compat.v1.Session() as sess:
#             sess.run(tf.compat.v1.global_variables_initializer())
#             for e in range(n_epoch):
#                 _, loss_ = sess.run([op, loss], feed_dict={x: X, y: Y})
#                 losses.append(loss_)
#                 if loss_ < tol:
#                     break
#                 if e > 0:
#                     losses_diff = abs(losses[-1] - losses[-2])
#                     if losses_diff < 1e-6:
#                         break
#             w_res = sess.run(w)
#             b_res = sess.run(b)
#         losses_final.append(losses)
#         W_final.append(w_res)
#         B_final.append(b_res)
#     W_final = np.concatenate(W_final, axis = 1)
#     B_final = np.concatenate(B_final, axis = 1)
#     if fit_intercept:
#         W_final = np.concatenate((W_final, B_final), axis = 0)
#     return W_final, losses_final



# -*- coding: utf-8 -*-

from numpy import ndarray
from typing import Union, Optional, Dict, List, Tuple
from scipy import signal

import numpy as np

from .algsupport import gen_ref_sin, floor

def ref_sig_fun(dataself, sig_len: float, N: int, phases: List[float], srate: float):
    """
    Because downsampling, the sampling rate of reference signals also should be changed
    """
    L = floor(sig_len * dataself.srate)
    ref_sig = [gen_ref_sin(freq, dataself.srate, L, N, phase) for freq, phase in zip(dataself.stim_info['freqs'], phases)]
    ref_sig = [signal.resample_poly(sig, 1000*srate, 1000*dataself.srate, axis = 1) for sig in ref_sig]
    return ref_sig

def suggested_weights_filterbank(num_subbands : int = 1) -> List[float]:
    if num_subbands == 1:
        return [1 for _ in range(num_subbands)]
    else:
        return [i**(-1.25)+0.25 for i in range(1,num_subbands+1,1)]

def suggested_ch() -> List[int]:
    """
    Provide suggested channels for openBMI

    Returns
    -------
    ch_used: List
        Suggested channels
    """
    return list(range(23-1,32))

def preprocess(dataself,
               X: ndarray,
               downsample_srate: float) -> ndarray:
    """
    Suggested preprocessing function for openBMI

    resample to 100 Hz
    """

    preprocess_X = signal.resample_poly(X, 1000*downsample_srate, 1000*dataself.srate, axis = 1)

    band = [0.5, 40]
    bpB, bpA = signal.butter(5, [band_val/(downsample_srate/2) for band_val in band], 'bandpass')
    tmp = signal.filtfilt(bpB, bpA, preprocess_X, axis = 1, padtype='odd', padlen=3*(max(len(bpB),len(bpA))-1))
    tmp = signal.detrend(tmp, axis = 1)
    tmp = tmp - np.mean(tmp, axis = 1, keepdims = True)
    tmp_std = np.std(tmp, axis = 1, keepdims = True, ddof=1)
    preprocess_X = tmp / tmp_std
    
    return preprocess_X

def filterbank(dataself,
               X: ndarray,
               srate: float,
               num_subbands: int = 1) -> ndarray:
    """
    Suggested filterbank function for openBMI
    """

    filterbank_X = np.zeros((num_subbands, X.shape[0], X.shape[1]))

    # band = [0.5, 40]
    # bpB, bpA = signal.butter(5, [band_val/(srate/2) for band_val in band], 'bandpass')
    # tmp = signal.filtfilt(bpB, bpA, X, axis = 1, padtype='odd', padlen=3*(max(len(bpB),len(bpA))-1))
    # tmp = signal.detrend(tmp, axis = 1)
    # tmp = tmp - np.mean(tmp, axis = 1, keepdims = True)
    # tmp_std = np.std(tmp, axis = 1, keepdims = True, ddof=1)
    # tmp = tmp / tmp_std
    # filterbank_X[0,:,:] = tmp.copy()
    
    for k in range(1, num_subbands+1, 1):
        if k == 1:
            filterbank_X[k-1,:,:] = X.copy()
        else:
            band = [60/11*k, 40]
            bpB, bpA = signal.butter(5, [band_val/(srate/2) for band_val in band], 'bandpass')
            tmp = signal.filtfilt(bpB, bpA, X, axis = 1, padtype='odd', padlen=3*(max(len(bpB),len(bpA))-1))
            tmp = signal.detrend(tmp, axis = 1)
            tmp = tmp - np.mean(tmp, axis = 1, keepdims = True)
            tmp_std = np.std(tmp, axis = 1, keepdims = True, ddof=1)
            filterbank_X[k-1,:,:] = tmp / tmp_std
            # Wp = [(60/11*k)/(srate/2), 80/(srate/2)]
            # Ws = [(60/11*k-1)/(srate/2), 90/(srate/2)]
            # N, Wn = signal.cheb1ord(Wp, Ws, 3, 40)
            
            # bpB, bpA = signal.cheby1(N, 0.5, Wn, btype = 'bandpass')
            # tmp = signal.filtfilt(bpB, bpA, X, axis = 1, padtype='odd', padlen=3*(max(len(bpB),len(bpA))-1))
            # filterbank_X[k-1,:,:] = signal.detrend(tmp, axis = 1)
        
    return filterbank_X




# -*- coding: utf-8 -*-

from typing import Union, Optional, Dict, List, Tuple, Callable
from numpy import ndarray

import numpy as np

def cal_confusionmatrix_onedataset_individual_online(evaluator: object,
                                                        dataset_idx: int,
                                                        tw_seq: List[float],
                                                        train_or_test: str) -> Tuple[ndarray, ndarray]:
    """
    Calculate confusion matrix for one dataset
    Evaluations will be carried out on each subject and each signal length
    Under each subject and each signal length, there may be several trials. Acc and itr values of all these trials will be averaged.

    Parameters
    ----------
    evaluator : object
        evaluator
    dataset_idx : int
        dataset index
    tw_seq : List[float]
        List of signal length
    train_or_test : str
        Calculate performance of training or testing performance

    Returns
    -------
    confusion_matrix: ndarray
        Classification accuracy
        Shape: (subject_num, signal_len_num, method_num, trial_num, sub_trial_num, true_class_num, pred_class_num)
    """
    if train_or_test.lower() == "train":
        idx = 0
    elif train_or_test.lower() == "test":
        idx = 1
    else:
        raise ValueError("Unknown train_or_test type. It must be 'train' or 'test'")
    dataset_container = evaluator.dataset_container
    model_container = evaluator.model_container
    sub_num = len(dataset_container[dataset_idx].subjects)
    N = dataset_container[dataset_idx].stim_info['stim_num']

    confusion_matrix = []
    for sub_idx in range(sub_num):
        tw_matrix = []
        for tw in tw_seq:

            trial_info = {'dataset_idx':[dataset_idx],
                          'sub_idx':[sub_idx],
                          'tw':tw}
            trial_idx = evaluator.search_trial_idx(train_or_test, trial_info)
            
            method_matrix = []
            for method_idx in range(len(model_container)):
                trial_matrix = []
                for i in trial_idx:
                    if idx == 0:
                        true_label_list = evaluator.performance_container[i][method_idx].true_label_train
                        pred_label_list = evaluator.performance_container[i][method_idx].pred_label_train
                    else:
                        true_label_list = evaluator.performance_container[i][method_idx].true_label_test
                        pred_label_list = evaluator.performance_container[i][method_idx].pred_label_test
                    sub_trial_matrix = []
                    for j in range(len(true_label_list)):
                        confusion_matrix_tmp = np.zeros((N,N))
                        for true_label, pred_label in zip(true_label_list[:(i+1)], pred_label_list[:(i+1)]):
                            confusion_matrix_tmp[true_label, pred_label] = confusion_matrix_tmp[true_label, pred_label] + 1
                        confusion_matrix_tmp = np.expand_dims(confusion_matrix_tmp, 0)
                        sub_trial_matrix.append(confusion_matrix_tmp)
                    sub_trial_matrix = np.concatenate(sub_trial_matrix, 0)
                    sub_trial_matrix = np.expand_dims(sub_trial_matrix, 0)
                    trial_matrix.append(sub_trial_matrix)
                trial_matrix = np.concatenate(trial_matrix, 0)
                trial_matrix = np.expand_dims(trial_matrix, 0)
                method_matrix.append(trial_matrix)
            method_matrix = np.concatenate(method_matrix, 0)
            method_matrix = np.expand_dims(method_matrix, 0)
            tw_matrix.append(method_matrix)
        tw_matrix = np.concatenate(tw_matrix, 0)
        tw_matrix = np.expand_dims(tw_matrix, 0)
        confusion_matrix.append(tw_matrix)
    confusion_matrix = np.concatenate(confusion_matrix, 0)
                    
    return confusion_matrix

def cal_performance_onedataset_individual_online(evaluator: object,
                                                     dataset_idx: int,
                                                     tw_seq: List[float],
                                                     train_or_test: str) -> Tuple[ndarray, ndarray]:
    """
    Calculate acc and ITR for one dataset
    Evaluations will be carried out on each subject and each signal length
    Under each subject and each signal length, there may be several trials. Acc and itr values of all these trials will be stored

    Parameters
    ----------
    evaluator : object
        evaluator
    dataset_idx : int
        dataset index
    tw_seq : List[float]
        List of signal length
    train_or_test : str
        Calculate performance of training or testing performance

    Returns
    -------
    acc_store: ndarray
        Classification accuracy
        Shape: (subject_num, signal_len_num, method_num, trial_num, sub_trial_num)
    itr_store: ndarray
        Classification ITR
        Shape: (subject_num, signal_len_num, method_num, trial_num, sub_trial_num)
    """
    if train_or_test.lower() == "train":
        idx = 0
    elif train_or_test.lower() == "test":
        idx = 1
    else:
        raise ValueError("Unknown train_or_test type. It must be 'train' or 'test'")
    dataset_container = evaluator.dataset_container
    model_container = evaluator.model_container
    sub_num = len(dataset_container[dataset_idx].subjects)
    t_break = dataset_container[dataset_idx].t_break

    # acc_store = np.zeros((len(model_container),sub_num, len(tw_seq)))
    # itr_store = np.zeros((len(model_container),sub_num, len(tw_seq)))
    acc_store = []
    itr_store = []

    for sub_idx in range(sub_num):
        acc_tw = []
        itr_tw = []
        for tw in tw_seq:
            trial_info = {'dataset_idx':[dataset_idx],
                          'sub_idx':[sub_idx],
                          'tw':tw}
            trial_idx = evaluator.search_trial_idx(train_or_test, trial_info)
            
            acc_method = []
            itr_method = []
            for method_idx in range(len(model_container)):
                tmp_acc = []
                tmp_itr = []
                for i in trial_idx:
                    performance_container = [evaluator.performance_container[i][method_idx]]
                    tmp_acc.extend(cal_acc_trials_onebyone(train_or_test, performance_container))
                    t_latency = evaluator.trial_container[i][idx].t_latency[dataset_idx]
                    tw = evaluator.trial_container[i][idx].tw
                    if t_latency is None:
                        t_latency = dataset_container[dataset_idx].default_t_latency
                    tmp_itr.extend(cal_itr_trials_onebyone(train_or_test, performance_container, tw, t_break, t_latency))
                tmp_acc = np.expand_dims(tmp_acc, 0)
                acc_method.append(tmp_acc)
                tmp_itr = np.expand_dims(tmp_itr, 0)
                itr_method.append(tmp_itr)
            acc_method = np.concatenate(acc_method, 0)
            acc_method = np.expand_dims(acc_method, 0)
            itr_method = np.concatenate(itr_method, 0)
            itr_method = np.expand_dims(itr_method, 0)
            acc_tw.append(acc_method)
            itr_tw.append(itr_method)
        acc_tw = np.concatenate(acc_tw, 0)
        acc_tw = np.expand_dims(acc_tw, 0)
        itr_tw = np.concatenate(itr_tw, 0)
        itr_tw = np.expand_dims(itr_tw, 0)
        acc_store.append(acc_tw)
        itr_store.append(itr_tw)
    acc_store = np.concatenate(acc_store, 0)
    itr_store = np.concatenate(itr_store, 0)

    return acc_store, itr_store

def cal_confusionmatrix_onedataset_individual_diffsiglen(evaluator: object,
                                                        dataset_idx: int,
                                                        tw_seq: List[float],
                                                        train_or_test: str,
                                                        subj_seq: Optional[List[int]] = None) -> Tuple[ndarray, ndarray]:
    """
    Calculate confusion matrix for one dataset
    Evaluations will be carried out on each subject and each signal length
    Under each subject and each signal length, there may be several trials. Acc and itr values of all these trials will be averaged.

    Parameters
    ----------
    evaluator : object
        evaluator
    dataset_idx : int
        dataset index
    tw_seq : List[float]
        List of signal length
    train_or_test : str
        Calculate performance of training or testing performance
    subj_seq : Optional[List[int]]
        List of subject indices
        If None, all subjects will be included

    Returns
    -------
    confusion_matrix: ndarray
        Classification accuracy
        Shape: (method_num, subject_num, signal_len_num, true_class_num, pred_class_num)
    """
    if train_or_test.lower() == "train":
        idx = 0
    elif train_or_test.lower() == "test":
        idx = 1
    else:
        raise ValueError("Unknown train_or_test type. It must be 'train' or 'test'")
    dataset_container = evaluator.dataset_container
    model_container = evaluator.model_container
    # sub_num = len(dataset_container[dataset_idx].subjects)
    N = dataset_container[dataset_idx].stim_info['stim_num']
    if subj_seq is None:
        subj_seq = list(range(len(dataset_container[dataset_idx].subjects)))
    sub_num = len(subj_seq)

    confusion_matrix = np.zeros((len(model_container),sub_num, len(tw_seq), N, N))

    for sub_i, sub_idx in enumerate(subj_seq):
        for tw_idx, tw in enumerate(tw_seq):
            trial_info = {'dataset_idx':[dataset_idx],
                          'sub_idx':[sub_idx],
                          'tw':tw}
            trial_idx = evaluator.search_trial_idx(train_or_test, trial_info)
            
            for method_idx in range(len(model_container)):
                for i in trial_idx:
                    if idx == 0:
                        true_label_list = evaluator.performance_container[i][method_idx].true_label_train
                        pred_label_list = evaluator.performance_container[i][method_idx].pred_label_train
                    else:
                        true_label_list = evaluator.performance_container[i][method_idx].true_label_test
                        pred_label_list = evaluator.performance_container[i][method_idx].pred_label_test
                    for true_label, pred_label in zip(true_label_list, pred_label_list):
                        confusion_matrix[method_idx, sub_i, tw_idx, true_label, pred_label] = confusion_matrix[method_idx, sub_i, tw_idx, true_label, pred_label] + 1
    return confusion_matrix

def cal_performance_onedataset_crossSubj_diffsiglen(evaluator: object,
                                                     dataset_idx: int,
                                                     tw_seq: List[float],
                                                     train_or_test: str) -> Tuple[ndarray, ndarray]:
    """
    Calculate acc and ITR for one dataset
    Evaluations will be carried out on each signal length
    Under each signal length, there may be several trials. Acc and itr values of all these trials will be averaged.

    Parameters
    ----------
    evaluator : object
        evaluator
    dataset_idx : int
        dataset index
    tw_seq : List[float]
        List of signal length
    train_or_test : str
        Calculate performance of training or testing performance

    Returns
    -------
    acc_store: ndarray
        Classification accuracy
        Shape: (method_num, subject_num, signal_len_num)
    itr_store: ndarray
        Classification ITR
        Shape: (method_num, subject_num, signal_len_num)
    """
    if train_or_test.lower() == "train":
        idx = 0
    elif train_or_test.lower() == "test":
        idx = 1
    else:
        raise ValueError("Unknown train_or_test type. It must be 'train' or 'test'")
    dataset_container = evaluator.dataset_container
    model_container = evaluator.model_container
    # sub_num = len(dataset_container[dataset_idx].subjects)
    t_break = dataset_container[dataset_idx].t_break
    
    acc_store = np.zeros((len(model_container),1, len(tw_seq)))
    itr_store = np.zeros((len(model_container),1, len(tw_seq)))
    for tw_idx, tw in enumerate(tw_seq):
        trial_info = {'dataset_idx':[dataset_idx],
                        'tw':tw}
        trial_idx = evaluator.search_trial_idx(train_or_test, trial_info)
        
        for method_idx in range(len(model_container)):
            tmp_acc = []
            tmp_itr = []
            for i in trial_idx:
                performance_container = [evaluator.performance_container[i][method_idx]]
                tmp_acc.append(cal_acc_trials(train_or_test, performance_container))
                t_latency = evaluator.trial_container[i][idx].t_latency[dataset_idx]
                tw = evaluator.trial_container[i][idx].tw
                if t_latency is None:
                    t_latency = dataset_container[dataset_idx].default_t_latency
                tmp_itr.append(cal_itr_trials(train_or_test, performance_container, tw, t_break, t_latency))
            acc_store[method_idx,0,tw_idx] = mean(tmp_acc)
            itr_store[method_idx,0,tw_idx] = mean(tmp_itr)
    return acc_store, itr_store

def cal_performance_onedataset_individual_diffsiglen(evaluator: object,
                                                     dataset_idx: int,
                                                     tw_seq: List[float],
                                                     train_or_test: str,
                                                     subj_seq: Optional[List[int]] = None) -> Tuple[ndarray, ndarray]:
    """
    Calculate acc and ITR for one dataset
    Evaluations will be carried out on each subject and each signal length
    Under each subject and each signal length, there may be several trials. Acc and itr values of all these trials will be averaged.

    Parameters
    ----------
    evaluator : object
        evaluator
    dataset_idx : int
        dataset index
    tw_seq : List[float]
        List of signal length
    train_or_test : str
        Calculate performance of training or testing performance
    subj_seq : Optional[List[int]]
        List of subject indices
        If None, all subjects will be included

    Returns
    -------
    acc_store: ndarray
        Classification accuracy
        Shape: (method_num, subject_num, signal_len_num)
    itr_store: ndarray
        Classification ITR
        Shape: (method_num, subject_num, signal_len_num)
    """
    if train_or_test.lower() == "train":
        idx = 0
    elif train_or_test.lower() == "test":
        idx = 1
    else:
        raise ValueError("Unknown train_or_test type. It must be 'train' or 'test'")
    dataset_container = evaluator.dataset_container
    model_container = evaluator.model_container
    # sub_num = len(dataset_container[dataset_idx].subjects)
    t_break = dataset_container[dataset_idx].t_break
    if subj_seq is None:
        subj_seq = list(range(len(dataset_container[dataset_idx].subjects)))
    sub_num = len(subj_seq)
    
    acc_store = np.zeros((len(model_container),sub_num, len(tw_seq)))
    itr_store = np.zeros((len(model_container),sub_num, len(tw_seq)))
    for sub_i, sub_idx in enumerate(subj_seq):
        for tw_idx, tw in enumerate(tw_seq):
            trial_info = {'dataset_idx':[dataset_idx],
                          'sub_idx':[sub_idx],
                          'tw':tw}
            trial_idx = evaluator.search_trial_idx(train_or_test, trial_info)
            
            for method_idx in range(len(model_container)):
                tmp_acc = []
                tmp_itr = []
                for i in trial_idx:
                    performance_container = [evaluator.performance_container[i][method_idx]]
                    tmp_acc.append(cal_acc_trials(train_or_test, performance_container))
                    t_latency = evaluator.trial_container[i][idx].t_latency[dataset_idx]
                    tw = evaluator.trial_container[i][idx].tw
                    if t_latency is None:
                        t_latency = dataset_container[dataset_idx].default_t_latency
                    tmp_itr.append(cal_itr_trials(train_or_test, performance_container, tw, t_break, t_latency))
                acc_store[method_idx,sub_i,tw_idx] = mean(tmp_acc)
                itr_store[method_idx,sub_i,tw_idx] = mean(tmp_itr)
    return acc_store, itr_store

def cal_itr_trials_onebyone(train_or_test: str,
                            performance_container: list,
                            tw,t_break,t_latency) -> List[float]:
    """
    Calculate itr of trials

    Parameters
    ----------
    train_or_test : str
        Test "train" or "test"
    performance_container : list
        List of performance

    Returns
    -------
    list_itr: List[float]
        List of ITR
    """
    list_acc = []
    for performance in performance_container:
        if train_or_test.lower() == "train":
            for i in range(len(performance.true_label_train)):
                acc = cal_acc(performance.true_label_train[:(i+1)], performance.pred_label_train[:(i+1)])
                t_comp = sum(performance.test_time_train[:(i+1)])/len(performance.true_label_train[:(i+1)])
                N = len(set(performance.true_label_train[:(i+1)]))
                list_acc.append(cal_itr(tw,t_break,t_latency,t_comp,N, acc))
        elif train_or_test.lower() == "test":
            for i in range(len(performance.true_label_test)):
                acc = cal_acc(performance.true_label_test[:(i+1)], performance.pred_label_test[:(i+1)])
                t_comp = sum(performance.test_time_test[:(i+1)])/len(performance.true_label_test[:(i+1)])
                N = len(set(performance.true_label_test[:(i+1)]))
                list_acc.append(cal_itr(tw,t_break,t_latency,t_comp,N, acc))
        else:
            raise ValueError("Unknown train_or_test type. It must be 'train' or 'test'")
    return list_acc

def cal_itr_trials(train_or_test: str,
                   performance_container: list,
                   tw,t_break,t_latency) -> float:
    """
    Calculate itr of trials

    Parameters
    ----------
    train_or_test : str
        Test "train" or "test"
    performance_container : list
        List of performance

    Returns
    -------
    itr: float
    """
    list_acc = []
    for performance in performance_container:
        if train_or_test.lower() == "train":
            acc = cal_acc(performance.true_label_train, performance.pred_label_train)
            t_comp = sum(performance.test_time_train)/len(performance.true_label_train)
            N = len(set(performance.true_label_train))
            list_acc.append(cal_itr(tw,t_break,t_latency,t_comp,N, acc))
        elif train_or_test.lower() == "test":
            acc = cal_acc(performance.true_label_test, performance.pred_label_test)
            t_comp = sum(performance.test_time_test)/len(performance.true_label_test)
            N = len(set(performance.true_label_test))
            list_acc.append(cal_itr(tw,t_break,t_latency,t_comp,N, acc))
        else:
            raise ValueError("Unknown train_or_test type. It must be 'train' or 'test'")
            
    return mean(list_acc)

def cal_acc_trials_onebyone(train_or_test: str,
                            performance_container: list) -> List[float]:
    """
    Calculate acc of trials

    Parameters
    ----------
    train_or_test : str
        Test "train" or "test"
    performance_container : list
        List of performance

    Returns
    -------
    list_acc: List[float]
        List of accuracy
    """
    list_acc = []
    for performance in performance_container:
        if train_or_test.lower() == "train":
            for i in range(len(performance.true_label_train)):
                list_acc.append(cal_acc(performance.true_label_train[:(i+1)], performance.pred_label_train[:(i+1)]))
        elif train_or_test.lower() == "test":
            for i in range(len(performance.true_label_test)):
                list_acc.append(cal_acc(performance.true_label_test[:(i+1)], performance.pred_label_test[:(i+1)]))
        else:
            raise ValueError("Unknown train_or_test type. It must be 'train' or 'test'")
    return list_acc

def cal_acc_trials(train_or_test: str,
                   performance_container: list) -> float:
    """
    Calculate acc of trials

    Parameters
    ----------
    train_or_test : str
        Test "train" or "test"
    performance_container : list
        List of performance

    Returns
    -------
    acc: float
    """
    list_acc = []
    for performance in performance_container:
        if train_or_test.lower() == "train":
            list_acc.append(cal_acc(performance.true_label_train, performance.pred_label_train))
        elif train_or_test.lower() == "test":
            list_acc.append(cal_acc(performance.true_label_test, performance.pred_label_test))
        else:
            raise ValueError("Unknown train_or_test type. It must be 'train' or 'test'")
            
    return mean(list_acc)
    

def mean(X:list) -> float:
    """
    Calculate mean of a list

    Parameters
    ----------
    X : list

    Returns
    -------
    mean : float

    """    
    return sum(X) / len(X)

def cal_acc(Y_true: List[int],
            Y_pred: List[int]) -> float:
    """
    Calculate accuracy

    Parameters
    ----------
    Y_true : List[int]
        True labels
    Y_pred : List[int]
        Predicted labels

    Returns
    -------
    acc : float
        Accuracy
    """
    if len(Y_true) != len(Y_pred):
        raise ValueError('Lengths of true labels and predicted labels should be same')
    true_detect = [1 for i in range(len(Y_true)) if int(Y_true[i])==int(Y_pred[i])]
    true_detect_count = sum(true_detect)
    acc = true_detect_count/len(Y_true)
    return acc

def cal_itr(tw: Union[int, float],
            t_break: Union[int, float],
            t_latency: Union[int, float],
            t_comp: Union[int, float],
            N: int,
            acc: float) -> float:
    """
    Calculate ITR

    Parameters
    ----------
    tw : Union[int, float]
        Signal length (in second)
    t_break : Union[int, float]
        Time required for shifting visual attention
    t_latency : Union[int, float]
        Latency time
    t_comp : Union[int, float]
        Computational time of each trial
    N : int
        Number of classes
    acc : float
        Accuracy

    Returns
    -------
    ITR: float
        bits/min
    """
    total_t = tw + t_break + t_latency + t_comp
    if acc == 1:
        itr = 60/total_t * (np.log2(N) + acc * np.log2(acc))
    elif acc < 1/N:
        itr = 0
    else:
        itr = 60/total_t * (np.log2(N) + acc * np.log2(acc) + (1-acc) * np.log2( (1-acc)/(N-1) ))
    return float(itr)
        


import setuptools

with open("README.md", "r", encoding="utf8") as fh:
    long_description = fh.read()

setuptools.setup(
    name="SSVEPAnalysisToolbox",
    version="0.0.2",
    author="Ze Wang",
    author_email="pikipityw@gmail.com",
    description="Python package for SSVEP datasets and algorithms",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/pikipity/SSVEP-Analysis-Toolbox",
    packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    # Required packages
    install_requires=[
        'numpy<=1.23,>1.0',
        'scipy<=1.13,>1.0',
        'py7zr<=0.21',
        'pooch<=1.8,>1.0',
        'joblib<=1.4,>1.0',
        'matplotlib<=3.8,>3.0',
        'tqdm<=4.66,>4.0',
        'scikit-learn<=1.3,>1.0',
        'mat73<=1.0'
    ],
    python_requires='>=3.9',
)


# -*- coding: utf-8 -*-

import os
import numpy as np
import zipfile
import shutil

from typing import Union, Optional, Dict, List, Tuple
from numpy import ndarray, transpose

from .basedataset import BaseDataset
from .subjectinfo import SubInfo
from ..utils.download import download_single_file
from ..utils.io import loadmat

import warnings

class WearableDataset_wet(BaseDataset):
    """
    Wearable SSVEP Dataset (wet)

    This study relied on the BCI Brain-Controlled Robot Contest at the 2020 World Robot Contest to recruit participants.

    One hundred and two healthy subjects (64 males and 38 females, with an average age of 30.03 ± 0.79 years ranging from 8 to 52 years) with normal or corrected-to-normal eyesight participated in the experiment. total, 53 subjects wore the dry-electrode headband first and 49 subjects wore the wet-electrode headband first. 

    This research designed an online BCI system with a 12-target speller as a virtual keypad of a phone.

    An 8-channel NeuSenW (Neuracle, Ltd. Changzhou, China) wireless EEG acquisition system was used to record the SSVEPs in this study.

    Each block included 12 trials, and each trial corresponded to each target.

    EEG data were recorded using Neuracle EEG Recorder NeuSen W (Neuracle, Ltd.), a wireless EEG acquisition system with a sampling rate of 1000 Hz. Eight electrodes (POz, PO3, PO4, PO5, PO6, Oz, O1 and O2, sorted by channel index in the dataset) were placed at the parietal and occipital regions on the basis of the international 10 to 20 system to record SSVEPs and two electrodes were placed at the forehead as the reference and ground, respectively.

    In accordance with the stimulus onsets recorded in the event channel of the continuous EEG data, data epochs could be extracted. The length of each data epoch was 2.84 s, including 0.5 s before the stimulus onset, 0.14 s for visual response delay, 2 s for stimulus, and 0.2 s after stimulus. With the purpose of reducing the storage and computation costs, all data were down sampled to 250 Hz.

    The electrode impedances recorded before each block were provided in the data matrix of ‘Impedance.mat’ with dimensions of [8, 10, 2, 102]. The channel index are corresponding to POz, PO3, PO4, PO5, PO6, Oz, O1, O2. The numbers in the four dimensions represent the number of channels, blocks, headband types (1: wet, 2: dry) and subjects respectively. The impedance information can be used to study the relationship be tween impedance and BCI performance.

    The “Subjects_information.mat” file lists the information of all 102 subjects together with aquestionnaire on the comfort level and preference of the two headbands after the experiment. For each participant, there are 10 columns of parameters (factors). The first 4 colu mns are the subjects’ personal information including “subject index”, “gender”, “age”, and “dominant hand”. The 6 columns(5th 10th) are listed as results in questionnaires, which are “Comfort of dry electrode headband”, “Wearing time of dry electrode when pain occurs”, “Comfort of wet electrode headband”, “Wearing time of wet electrode when pain occurs”, “Only consider comfort, headband preference” and “comprehensively consider comfort and convenience (need assistance from others, conductive paste, shampoo, etc.), headband preference". The last column shows the order of wearing the two headbands.

    The “stimulation_information.pdf” file lists the stimulation parameters of the 12 characters, including frequency and phase information of each character.

    Total: around 929 MB

    Reference:
    F. Zhu, L. Jiang, G. Dong, X. Gao, and Y. Wang, “An Open Dataset for Wearable SSVEP-Based Brain-Computer Interfaces,” Sensors, vol. 21, no. 4, p. 1256, 2021. DOI: 10.3390/s21041256
    https://www.mdpi.com/1424-8220/21/4/1256
    """

    _CHANNELS = [
        'POz', 'PO3', 'PO4', 'PO5', 'PO6', 'Oz', 'O1', 'O2'
    ]

    _FREQS = [
        9.25, 11.25, 13.25, 9.75, 
        11.75, 13.75, 10.25, 12.25, 
        14.25, 10.75, 12.75, 14.75
    ]

    _PHASES = [
        0, 0, 0, 0.5, 
        0.5, 0.5, 1, 1, 
        1, 1.5, 1.5, 1.5
    ]

    _SUBJECTS = [SubInfo(ID = 'S{:d}'.format(sub_idx)) for sub_idx in range(1,102+1,1)]

    def __init__(self, 
                 path: Optional[str] = None,
                 path_support_file: Optional[str] = None,
                 ID = 'Wearable (wet)'):
        super().__init__(subjects = self._SUBJECTS, 
                         ID = ID, 
                         url = 'http://bci.med.tsinghua.edu.cn/upload/zhufangkun/', 
                         paths = path, 
                         channels = self._CHANNELS, 
                         srate = 250, 
                         block_num = 10, 
                         trial_num = len(self._FREQS),
                         trial_len = 2.84, 
                         stim_info = {'stim_num': len(self._FREQS),
                                      'freqs': self._FREQS,
                                      'phases': [i * np.pi for i in self._PHASES]},
                         support_files = ['Readme.pdf',
                                          'stimulation_information.pdf',
                                          'Subjects_Information.mat',
                                          'Impedance.mat'],
                         path_support_file = path_support_file,
                         t_prestim = 0.5,
                         t_break = 0.2,
                         default_t_latency = 0.14)
        self.trial_label_check_list = {}
        for trial_i in range(self.trial_num):
            self.trial_label_check_list[trial_i] = trial_i
    
    def download_single_subject(self,
                                subject: SubInfo):
        sub_idx = int(subject.ID[1:])
        if sub_idx < 10:
            file_name = 'S00{:n}.mat'.format(sub_idx)
        elif sub_idx < 100:
            file_name = 'S0{:n}.mat'.format(sub_idx)
        else:
            file_name = 'S{:n}.mat'.format(sub_idx)
        data_file = os.path.join(subject.path, file_name)

        sub_idx = int(subject.ID[1:])
        if sub_idx <= 10:
            file_name = 'S001-S010.zip'
        elif sub_idx <= 20:
            file_name = 'S011-S020.zip'
        elif sub_idx <= 30:
            file_name = 'S021-S030.zip'
        elif sub_idx <= 40:
            file_name = 'S031-S040.zip'
        elif sub_idx <= 50:
            file_name = 'S041-S050.zip'
        elif sub_idx <= 60:
            file_name = 'S051-S060.zip'
        elif sub_idx <= 70:
            file_name = 'S061-S070.zip'
        elif sub_idx <= 80:
            file_name = 'S071-S080.zip'
        elif sub_idx <= 90:
            file_name = 'S081-S090.zip'
        else:
            file_name = 'S091-S102.zip'
        source_url = self.url + file_name
        desertation = os.path.join(subject.path, file_name)

        download_flag = True

        if not os.path.isfile(data_file):
            try:
                download_single_file(source_url, desertation)

                with zipfile.ZipFile(desertation,'r') as archive:
                    archive.extractall(subject.path)

                os.remove(desertation)
            except:
                download_flag = False

        return download_flag, source_url, desertation
    
    def download_file(self,
                      file_name: str):
        source_url = self.url + file_name
        desertation = os.path.join(self.path_support_file, file_name)

        download_flag = True
        
        if not os.path.isfile(desertation):
            try:
                download_single_file(source_url, desertation)
            except:
                download_flag = False

        return download_flag, source_url, desertation

    def download_support_files(self, total_retry_time = 10):
        """
        Download all support files
        """
        if self.support_files is not None and self.path_support_file is not None:
            for file_name in self.support_files:
                download_try_count = 0
                download_flag = False
                while (not download_flag) and (download_try_count < total_retry_time):
                    if download_try_count>0:
                        warnings.warn("There is an error when donwloading '{:s}'. So retry ({:n})".format(file_name, download_try_count))
                    download_flag, source_url, desertation = self.download_file(file_name)
                    if (not download_flag):
                        if os.path.isfile(desertation):
                            os.remove(desertation)
                    download_try_count += 1
                # load subject information
                if download_flag:
                    if file_name == 'Subjects_Information.mat':
                        sub_mat = np.array(loadmat(desertation)['Subjects_Information'], dtype=object)
                        for sub_idx in range(len(self.subjects)):
                            self.subjects[sub_idx].age = sub_mat[sub_idx+1, 2]
                            if sub_mat[sub_idx+1, 1] == 'Male':
                                self.subjects[sub_idx].gender = 'M'
                            else:
                                self.subjects[sub_idx].gender = 'F'
                else:
                    raise ValueError("Cannot download '{:s}'.".format(file_name))

    def get_sub_data(self, 
                     sub_idx: int) -> ndarray:
        if sub_idx < 0:
            raise ValueError('Subject index cannot be negative')
        if sub_idx > len(self.subjects)-1:
            raise ValueError('Subject index should be smaller than {:d}'.format(len(self.subjects)))

        sub_info = self.subjects[sub_idx]
        sub_idx = int(sub_info.ID[1:])
        if sub_idx < 10:
            file_name = 'S00{:n}.mat'.format(sub_idx)
        elif sub_idx < 100:
            file_name = 'S0{:n}.mat'.format(sub_idx)
        else:
            file_name = 'S{:n}.mat'.format(sub_idx)
        file_path = os.path.join(sub_info.path, file_name)

        mat_data = loadmat(file_path)
        data = mat_data['data'][:,:,1,:,:] # Only wet
        data = transpose(data, (2,3,0,1))

        return data

    def get_label_single_trial(self,
                               sub_idx: int,
                               block_idx: int,
                               trial_idx: int) -> int:
        return self.trial_label_check_list[trial_idx]

class WearableDataset_dry(WearableDataset_wet):
    """
    Wearable SSVEP Dataset (wet)
    """
    def __init__(self, 
                 path: Optional[str] = None,
                 path_support_file: Optional[str] = None):
        super().__init__(path = path,
                        path_support_file = path_support_file,
                        ID = 'Wearable (dry)')
    def get_sub_data(self, 
                     sub_idx: int) -> ndarray:
        if sub_idx < 0:
            raise ValueError('Subject index cannot be negative')
        if sub_idx > len(self.subjects)-1:
            raise ValueError('Subject index should be smaller than {:d}'.format(len(self.subjects)))

        sub_info = self.subjects[sub_idx]
        sub_idx = int(sub_info.ID[1:])
        if sub_idx < 10:
            file_name = 'S00{:n}.mat'.format(sub_idx)
        elif sub_idx < 100:
            file_name = 'S0{:n}.mat'.format(sub_idx)
        else:
            file_name = 'S{:n}.mat'.format(sub_idx)
        file_path = os.path.join(sub_info.path, file_name)

        mat_data = loadmat(file_path)
        data = mat_data['data'][:,:,0,:,:] # Only dry
        data = transpose(data, (2,3,0,1))

        return data


# -*- coding: utf-8 -*-

import os
import numpy as np
import tarfile

from typing import Union, Optional, Dict, List, Tuple
from numpy import ndarray, transpose

from .basedataset import BaseDataset
from .subjectinfo import SubInfo
from ..utils.download import download_single_file
from ..utils.io import loadmat

class ELDBETADataset(BaseDataset):
    """
    eldBETA Dataset

    For the BCI users, there was an associated epoched record that is stored in ".mat" structure array from MATLAB. 
    
    The structure array in each record was composed of the EEG data ("EEG") and its associated supplementary information ("Suppl_info") as its fields. In the "EEG" field of the record, two types of EEG data, i.e., EEG epochs and raw EEG were provided for researchers to facilitate diverse research purposes. 
    
    The EEG epochs were the EEG data with the data processing and stored as 4-dimensional matrices (channel x time point x condition x block). The names and locations of the channel dimension were given in the supplementary information. 
    
    For the dimension of time point, the epochs had a length of 6 s, which included 0.5 s before the stimulus onset, 5 s during the stimulation (SSVEPs) and 0.5 s after the stimulus offset. 
    
    Different from the epoched data, the raw EEG provided continuous EEG that were converted by EEGLAB. The raw EEG were stored as cell arrays, each of which contained a block of EEG data. The "Suppl_info" field of the record provided a basic information about personal statistics and experimental protocol. The personal statistics included the aged, gender, BCIQ and SNR with respect to each subject. The experimental protocol included channel location ("Channel), stimulus frequency ("Frequency"), stimulus initial phase ("Phase") and sampling rate ("Srate"). The channel location was represented by a 64x4 cell arrays. The first column and the fourth column denoted the channel index and channel name, respectively. The second column and the third column denoted the channel location in polar coordinates, i.e., degree and radius, respectively. The stimulus initial phase was given in radius. The sampling rate of the epoch data was denoted by "Srate". 

    Total: around 20.0 GB

    Paper:
    Liu, B., Wang, Y., Gao, X. and Chen, X., 2022. eldBETA: A Large Eldercare-oriented Benchmark Database of SSVEP-BCI for the Aging Population. Scientific Data, 9(1), pp.1-12.

    -----------------Corrigendum------------------------------------------
    - 1. In page 4, "perspective" should be changed to "prospective".
    - 2. In figure 6 and the main text, the method name "temporally local multivariate synchronization index (tMSI)" should be changed to "an extension to MSI (EMSI)", in which we chose the best competing algorithm (i.e., EMSI) and the names were confused. The green line in figure 6 denotes EMSI in "Zhang, Y., Guo, D., Yao, D. and Xu, P., 2017. The extension of multivariate synchronization index method for SSVEP-based BCI. Neurocomputing, 269, pp.226-231".
    """

    _CHANNELS = [
        'FP1','FPZ','FP2','AF3','AF4','F7','F5','F3','F1','FZ','F2','F4','F6',
        'F8','FT7','FC5','FC3','FC1','FCZ','FC2','FC4','FC6','FT8','T7','C5',
        'C3','C1','CZ','C2','C4','C6','T8','M1','TP7','CP5','CP3','CP1','CPZ',
        'CP2','CP4','CP6','TP8','M2','P7','P5','P3','P1','PZ','P2','P4','P6',
        'P8','PO7','PO5','PO3','POZ','PO4','PO6','PO8','CB1','O1','OZ','O2','CB2'
    ]

    _FREQS = [
        8.0, 9.5, 11,
        8.5, 10, 11.5,
        9, 10.5, 12
    ]

    _PHASES = [
        0, 1.5, 1,
        0.5, 0, 1.5,
        1, 0.5, 0
    ]

    _SUBJECTS = [SubInfo(ID = 'S{:d}'.format(sub_idx)) for sub_idx in range(1,100+1,1)]

    def __init__(self, 
                 path: Optional[str] = None,
                 path_support_file: Optional[str] = None):
        super().__init__(subjects = self._SUBJECTS, 
                         ID = 'eldBETA Dataset', 
                         url = 'http://bci.med.tsinghua.edu.cn/upload/liubingchuan_eldBETA_database/', 
                         paths = path, 
                         channels = self._CHANNELS, 
                         srate = 250, 
                         block_num = 7, 
                         trial_num = len(self._FREQS),
                         trial_len = 6, 
                         stim_info = {'stim_num': len(self._FREQS),
                                      'freqs': self._FREQS,
                                      'phases': [i * np.pi for i in self._PHASES]},
                         support_files = ['Description.tar.gz'],
                         path_support_file = path_support_file,
                         t_prestim = 0.5,
                         t_break = 0.5,
                         default_t_latency = 0.14)
        self.trial_label_check_list = {}
        for trial_i in range(self.trial_num):
            self.trial_label_check_list[trial_i] = trial_i

    def download_single_subject(self,
                                subject: SubInfo):
        data_file = os.path.join(subject.path, subject.ID + '.mat')

        sub_idx = int(subject.ID[1:])
        if sub_idx <= 10:
            file_name = 'S1-S10.tar.gz'
        elif sub_idx <= 20:
            file_name = 'S11-S20.tar.gz'
        elif sub_idx <= 30:
            file_name = 'S21-S30.tar.gz'
        elif sub_idx <= 40:
            file_name = 'S31-S40.tar.gz'
        elif sub_idx <= 50:
            file_name = 'S41-S50.tar.gz'
        elif sub_idx <= 60:
            file_name = 'S51-S60.tar.gz'
        elif sub_idx <= 70:
            file_name = 'S61-S70.tar.gz'
        elif sub_idx <= 80:
            file_name = 'S71-S80.tar.gz'
        elif sub_idx <= 90:
            file_name = 'S81-S90.tar.gz'
        elif sub_idx <= 100:
            file_name = 'S91-S100.tar.gz'
        source_url = self.url + file_name
        desertation = os.path.join(subject.path, file_name)

        download_flag = True

        if not os.path.isfile(data_file):
            try:
                download_single_file(source_url, desertation)

                with tarfile.open(desertation,'r') as archive:
                    archive.extractall(subject.path)
                
                os.remove(desertation)
            except:
                download_flag = False

        if download_flag:
            # load age and gender
            mat_data = loadmat(data_file)['data']['Suppl_info']
            age = mat_data['Age']
            gender = mat_data['Gender']
            if gender.lower()=='male':
                gender = 'M'
            else:
                gender = 'F'
            
            for sub_idx, sub_info in enumerate(self.subjects):
                if subject.ID == sub_info.ID:
                    self.subjects[sub_idx].age = age
                    self.subjects[sub_idx].gender = gender
                    break

        return download_flag, source_url, desertation
    
    def download_file(self,
                      file_name: str):
        source_url = self.url + file_name
        desertation = os.path.join(self.path_support_file, file_name)

        download_flag = True
        
        if not os.path.isfile(desertation):
            try:
                download_single_file(source_url, desertation)
            except:
                download_flag = False
        
        return download_flag, source_url, desertation

    def get_sub_data(self, 
                     sub_idx: int) -> ndarray:
        if sub_idx < 0:
            raise ValueError('Subject index cannot be negative')
        if sub_idx > len(self.subjects)-1:
            raise ValueError('Subject index should be smaller than {:d}'.format(len(self.subjects)))
        
        sub_info = self.subjects[sub_idx]
        file_path = os.path.join(sub_info.path, sub_info.ID + '.mat')
        
        mat_data = loadmat(file_path)
        data = mat_data['data']['EEG']['Epoch']
        data = transpose(data, (3,2,0,1)) # block_num * stimulus_num * ch_num * whole_trial_samples

        return data

    def get_label_single_trial(self,
                               sub_idx: int,
                               block_idx: int,
                               trial_idx: int) -> int:
        return self.trial_label_check_list[trial_idx]
        


# -*- coding: utf-8 -*-
"""
MS-TRCA-R-1
"""

from typing import Union, Optional, Dict, List, Tuple, Callable
from numpy import ndarray
from joblib import Parallel, delayed
from functools import partial
from copy import deepcopy

import numpy as np

from .lsframework import TDCA_ls
from .utils_ls import (
    get_lsconfig,
    ssvep_lsframe
)
from .utils import (
    qr_list, sort, separate_trainSig
)
from .tdca import (
    _gen_delay_X
)

class TRCAwithR_multi_f(TDCA_ls):
    def __init__(self,
                 n_neighbor: int = 2,
                 n_component: int = 1,
                 n_jobs: Optional[int] = None,
                 weights_filterbank: Optional[List[float]] = None,
                 n_delay: int = 0,
                 LSMethod = 'lstsq',
                 displayConvergWarn: bool = False,
                 max_n: int = 100,
                 M_error_threshold: float = 1e-6,
                 W_error_threshold: float = 1e-6,
                 l1_alpha = None,
                 l2_alpha = None,
                 alpha = None,
                 reg_iter = 1000,
                 reg_tol = 1e-4):
        super().__init__(n_component, n_jobs, weights_filterbank, n_delay, LSMethod, displayConvergWarn, max_n, M_error_threshold, W_error_threshold, l1_alpha, l2_alpha, alpha, reg_iter, reg_tol)
        self.ID = 'TRCAwithR_multi_f'
        self.n_neighbor = n_neighbor

    def __copy__(self):
        copy_model = TRCAwithR_multi_f(n_neighbor = self.n_neighbor,
                                    n_component = self.n_component,
                                    n_jobs = self.n_jobs,
                                    weights_filterbank = self.model['weights_filterbank'],
                                    n_delay = self.n_delay)
        copy_model.LSconfig = deepcopy(self.LSconfig)
        copy_model.model = deepcopy(self.model)
        return copy_model
    
    def fit(self,
            X: Optional[List[ndarray]] = None,
            Y: Optional[List[int]] = None,
            ref_sig: Optional[List[ndarray]] = None,
            freqs: Optional[List[float]] = None,
            *argv, **kwargv):
        """
        Parameters
        -------------
        X : Optional[List[ndarray]], optional
            List of training EEG data. The default is None.
            List shape: (trial_num,)
            EEG shape: (filterbank_num, channel_num, signal_len)
        Y : Optional[List[int]], optional
            List of labels (stimulus indices). The default is None.
            List shape: (trial_num,)
        ref_sig : Optional[List[ndarray]], optional
            Sine-cosine-based reference signals. The default is None.
            List of shape: (stimulus_num,)
            Reference signal shape: (harmonic_num, signal_len)
        """
        if freqs is None:
            raise ValueError('ms-eTRCA requires the list of stimulus frequencies')
        if Y is None:
            raise ValueError('TDCA requires training label')
        if X is None:
            raise ValueError('TDCA requires training data')
        if ref_sig is None:
            raise ValueError("TDCA requires reference signals")

        ref_sig_Q, _, _ = qr_list(ref_sig)
        ref_sig_P = [Q @ Q.T for Q in ref_sig_Q]
        self.model['ref_sig_P'] = ref_sig_P

        n_delay = self.n_delay
        X_P = _gen_delay_X(X, n_delay)
        separated_trainSig_1 = separate_trainSig(X_P, Y)
        separated_trainSig = [np.concatenate([trainSig_single_freq @ P_single_freq], axis=-1)
                              for trainSig_single_freq, P_single_freq in zip(separated_trainSig_1, ref_sig_P)]

        # template signals and spatial filters
        #   U: (filterbank_num * stimulus_num * channel_num * n_component)
        #   X or template_sig: (filterbank_num, channel_num, signal_len)
        filterbank_num, channel_num, signal_len = X_P[0].shape
        stimulus_num = len(ref_sig)
        n_component = self.n_component
        n_neighbor = self.n_neighbor
        d0 = int(np.floor(n_neighbor/2))
        _, freqs_idx, return_freqs_idx = sort(freqs)
        separated_trainSig = [separated_trainSig[i] for i in freqs_idx]
        separated_trainSig_1 = [separated_trainSig_1[i] for i in freqs_idx]
        # possible_class = list(set(Y))
        # possible_class.sort(reverse = False)
        # template_sig = [np.zeros((filterbank_num, channel_num, signal_len)) for _ in range(stimulus_num)]
        U_tdca = np.zeros((filterbank_num, stimulus_num, channel_num, n_component))
        for filterbank_idx in range(filterbank_num):
            for class_idx in range(1,stimulus_num+1):
                if class_idx <= d0:
                    start_idx = 0
                    end_idx = n_neighbor
                elif class_idx > d0 and class_idx < (stimulus_num-d0+1):
                    start_idx = class_idx - d0 - 1
                    end_idx = class_idx + (n_neighbor-d0-1)
                else:
                    start_idx = stimulus_num - n_neighbor
                    end_idx = stimulus_num
                separated_trainSig_tmp = [separated_trainSig[i] for i in range(start_idx, end_idx)]
                # Calulcate spatial filter
                W = _TDCA_ls_calW([np.swapaxes(trainSig[:,filterbank_idx,:,:], 1,2) for trainSig in separated_trainSig_tmp],
                                self.n_jobs,
                                LSconfig = self.LSconfig)
                U_tdca[filterbank_idx,class_idx-1,:,:] = W[:,:n_component]
        # Calculate template
        separated_trainSig_1 = [separated_trainSig_1[i] for i in return_freqs_idx]
        template_sig = [np.mean(tmp, 0) for tmp in separated_trainSig_1]
        # Generate spatial filters for all stimuli
        # U_tdca = np.repeat(U_tdca[:,:,:,return_freqs_idx], repeats = stimulus_num, axis = 1)
        self.model['U'] = U_tdca[:,return_freqs_idx,:,:]
        self.model['template_sig'] = template_sig
    
    def predict(self,
            X: List[ndarray]) -> List[int]:
        weights_filterbank = self.model['weights_filterbank']
        if weights_filterbank is None:
            weights_filterbank = [1 for _ in range(X[0].shape[0])]
        if type(weights_filterbank) is list:
            weights_filterbank = np.expand_dims(np.array(weights_filterbank),1).T
        else:
            if len(weights_filterbank.shape) != 2:
                raise ValueError("'weights_filterbank' has wrong shape")
            if weights_filterbank.shape[0] != 1:
                weights_filterbank = weights_filterbank.T
        if weights_filterbank.shape[0] != 1:
            raise ValueError("'weights_filterbank' has wrong shape")
        n_delay = self.n_delay

        X_delay = _gen_delay_X(X, n_delay)

        template_sig = self.model['template_sig']
        U = self.model['U'] 
        ref_sig_P = self.model['ref_sig_P']

        if self.n_jobs is not None:
            r = Parallel(n_jobs=self.n_jobs)(delayed(partial(_r_tdca_canoncorr_withUV, Y=template_sig, P=ref_sig_P, U=U, V=U))(X=a) for a in X_delay)
        else:
            r = []
            for a in X_delay:
                r.append(
                    _r_tdca_canoncorr_withUV(X=a, Y=template_sig, P=ref_sig_P, U=U, V=U)
                )

        Y_pred = [int( np.argmax( weights_filterbank @ r_tmp)) for r_tmp in r]
        
        return Y_pred, r
    

def _TDCA_ls_calW(trainSig, n_jobs = None, LSconfig = None):
    def X_single_freq(trainSig_single_freq, All_train_sum):
        Nt, Np, Nc = trainSig_single_freq.shape
        X = []
        for _ in range(Nt):
            X_tmp = np.sum(trainSig_single_freq, axis = 0) - 1/Nf * All_train_sum
            X.append(X_tmp)
        X = np.concatenate(X, axis = 0)
        return X
    def Y_single_freq(trainSig_single_freq):
        Nt, Np, Nc = trainSig_single_freq.shape
        Y = []
        for n in range(Nt):
            Y_tmp = trainSig_single_freq[n,:,:] - 1/Nt * np.sum(trainSig_single_freq, axis = 0)
            Y.append(Y_tmp)
        Y = np.concatenate(Y, axis = 0)
        return Y
    #
    LSMethod, displayConvergWarn, max_n, M_error_threshold, W_error_threshold, l1_alpha, l2_alpha, alpha, reg_iter, reg_tol, check_full_rank, kernel_fun = get_lsconfig(LSconfig)
    #
    Nf = len(trainSig)
    All_train_sum = None
    for trainSig_single_freq in trainSig:
        if All_train_sum is None:
            All_train_sum = np.sum(trainSig_single_freq, axis = 0)
        else:
            All_train_sum = All_train_sum + np.sum(trainSig_single_freq, axis = 0)
    if n_jobs is None:
        X = []
        for trainSig_single_freq in trainSig:
            X.append(
                X_single_freq(trainSig_single_freq, All_train_sum)
            )
            # Nt, Np, Nc = trainSig_single_freq.shape
            # for n in range(Nt):
            #     X_tmp = np.sum(trainSig_single_freq, axis = 0) - 1/Nf * All_train_sum
            #     X.append(X_tmp)
    else:
        X = Parallel(n_jobs=n_jobs)(delayed(partial(X_single_freq, All_train_sum = All_train_sum))
                                                   (trainSig_single_freq = trainSig_single_freq)
                                                   for trainSig_single_freq in trainSig)
    X = np.concatenate(X, axis = 0)
    if n_jobs is None:
        Y = []
        for trainSig_single_freq in trainSig:
            Y.append(
                Y_single_freq(trainSig_single_freq)
            )
            # Nt, Np, Nc = trainSig_single_freq.shape
            # for n in range(Nt):
            #     Y_tmp = trainSig_single_freq[n,:,:] - 1/Nt * np.sum(trainSig_single_freq, axis = 0)
            #     Y.append(Y_tmp)
    else:
        Y = Parallel(n_jobs=n_jobs)(delayed(Y_single_freq)
                                            (trainSig_single_freq = trainSig_single_freq)
                                            for trainSig_single_freq in trainSig)
    Y = np.concatenate(Y, axis = 0)
    W, M, M_error, W_error, ls_n, full_rank_check = ssvep_lsframe(X, Y,
                                                                  LSMethod = LSMethod,
                                                                  displayConvergWarn = displayConvergWarn,
                                                                  max_n = max_n,
                                                                  M_error_threshold = M_error_threshold,
                                                                  W_error_threshold = W_error_threshold,
                                                                  largest_n = max_n,
                                                                  check_rank = check_full_rank,
                                                                  l1_alpha = l1_alpha,
                                                                  l2_alpha = l2_alpha,
                                                                  alpha = alpha,
                                                                  reg_iter = reg_iter,
                                                                  reg_tol = reg_tol,
                                                                  kernel_fun = kernel_fun)
    #
    # Nf = len(trainSig)
    # Z = np.ndarray([])
    # Lx_first = np.ndarray([])
    # # Px = np.ndarray([])
    # Nt_total = 0
    # for trainSig_single_freq in trainSig:
    #     Nt, Np, Nc = trainSig_single_freq.shape
    #     # Q_single_freq, _, _ = slin.qr(ref_sig[Np_total:(Np_total + Np),:], mode = 'economic', pivoting = True)
    #     # Np_total += Np
    #     # _, Nh = Q_single_freq.shape
    #     Nt_total += Nt
    #     if len(Z.shape)==0:
    #         Z = blkmat(trainSig_single_freq)
    #         Lx_first = column_eyes(Nt, Np) @ column_eyes(Nt, Np).T
    #         # Px = blkrep(np.concatenate([np.eye(Np),P_single_freq],axis=1), Nt)
    #     else:
    #         Z = slin.block_diag(Z, blkmat(trainSig_single_freq))
    #         Lx_first = slin.block_diag(Lx_first, column_eyes(Nt, Np) @ column_eyes(Nt, Np).T)
    #         # Px = slin.block_diag(Px, blkrep(np.concatenate([np.eye(Np),P_single_freq],axis=1), Nt))
    # Z = Z @ column_eyes(Nt_total, Nc)
    # Lx = Lx_first - 1/Nf * (column_eyes(Nt_total, Np) @ column_eyes(Nt_total, Np).T)
    # Px = np.eye(Nt_total * Np)
    # Ly = np.eye(Nt_total * Np) - 1/Nt * Lx_first
    # W, M, M_error, W_error, ls_n, full_rank_check = ssvep_lsframe(Z, Lx, Px, Ly, Px)
    #
    W = W[-1].copy()
    return W

def _r_tdca_canoncorr_withUV(X: ndarray,
                            Y: List[ndarray],
                            P: List[ndarray],
                            U: ndarray,
                            V: ndarray) -> ndarray:
    """
    Calculate correlation of CCA based on canoncorr for single trial data using existing U and V

    Parameters
    ----------
    X : ndarray
        Single trial EEG data
        EEG shape: (filterbank_num, channel_num, signal_len)
    Y : List[ndarray]
        List of reference signals
    P : List[ndarray]
        List of P
    U : ndarray
        Spatial filter
        shape: (filterbank_num * stimulus_num * channel_num * n_component)
    V : ndarray
        Weights of harmonics
        shape: (filterbank_num * stimulus_num * harmonic_num * n_component)

    Returns
    -------
    R : ndarray
        Correlation
        shape: (filterbank_num * stimulus_num)
    """
    filterbank_num, channel_num, signal_len = X.shape
    if len(Y[0].shape)==2:
        harmonic_num = Y[0].shape[0]
    elif len(Y[0].shape)==3:
        harmonic_num = Y[0].shape[1]
    else:
        raise ValueError('Unknown data type')
    stimulus_num = len(Y)
    
    R = np.zeros((filterbank_num, stimulus_num))
    
    for k in range(filterbank_num):
        tmp = X[k,:,:]
        for i in range(stimulus_num):
            # tmp = np.concatenate([tmp_X, tmp_X @ P[i]], axis=-1)
            if len(Y[i].shape)==2:
                Y_tmp = Y[i]
            elif len(Y[i].shape)==3:
                Y_tmp = Y[i][k,:,:]
            else:
                raise ValueError('Unknown data type')
            
            A_r = U[k,i,:,:]
            B_r = V[k,i,:,:]
            
            a = A_r.T @ tmp
            b = B_r.T @ Y_tmp
            a = np.reshape(a, (-1))
            b = np.reshape(b, (-1))
            
            # r2 = stats.pearsonr(a, b)[0]
            # r = stats.pearsonr(a, b)[0]
            r = np.corrcoef(a, b)[0,1]
            R[k,i] = r
    return R


class eTRCAwithR_multi_f(TDCA_ls):
    def __init__(self,
                 n_neighbor: int = 2,
                 n_component: int = 1,
                 n_jobs: Optional[int] = None,
                 weights_filterbank: Optional[List[float]] = None,
                 n_delay: int = 0,
                 LSMethod = 'lstsq',
                 displayConvergWarn: bool = False,
                 max_n: int = 100,
                 M_error_threshold: float = 1e-6,
                 W_error_threshold: float = 1e-6,
                 l1_alpha = None,
                 l2_alpha = None,
                 alpha = None,
                 reg_iter = 1000,
                 reg_tol = 1e-4):
        super().__init__(n_component, n_jobs, weights_filterbank, n_delay, LSMethod, displayConvergWarn, max_n, M_error_threshold, W_error_threshold, l1_alpha, l2_alpha, alpha, reg_iter, reg_tol)
        self.ID = 'eTRCAwithR_multi_f'
        self.n_neighbor = n_neighbor

    def __copy__(self):
        copy_model = eTRCAwithR_multi_f(n_neighbor = self.n_neighbor,
                                    n_component = self.n_component,
                                    n_jobs = self.n_jobs,
                                    weights_filterbank = self.model['weights_filterbank'],
                                    n_delay = self.n_delay)
        copy_model.LSconfig = deepcopy(self.LSconfig)
        copy_model.model = deepcopy(self.model)
        return copy_model
    
    def fit(self,
            X: Optional[List[ndarray]] = None,
            Y: Optional[List[int]] = None,
            ref_sig: Optional[List[ndarray]] = None,
            freqs: Optional[List[float]] = None,
            *argv, **kwargv):
        """
        Parameters
        -------------
        X : Optional[List[ndarray]], optional
            List of training EEG data. The default is None.
            List shape: (trial_num,)
            EEG shape: (filterbank_num, channel_num, signal_len)
        Y : Optional[List[int]], optional
            List of labels (stimulus indices). The default is None.
            List shape: (trial_num,)
        ref_sig : Optional[List[ndarray]], optional
            Sine-cosine-based reference signals. The default is None.
            List of shape: (stimulus_num,)
            Reference signal shape: (harmonic_num, signal_len)
        """
        if freqs is None:
            raise ValueError('ms-eTRCA requires the list of stimulus frequencies')
        if Y is None:
            raise ValueError('TDCA requires training label')
        if X is None:
            raise ValueError('TDCA requires training data')
        if ref_sig is None:
            raise ValueError("TDCA requires reference signals")

        ref_sig_Q, _, _ = qr_list(ref_sig)
        ref_sig_P = [Q @ Q.T for Q in ref_sig_Q]
        self.model['ref_sig_P'] = ref_sig_P

        n_delay = self.n_delay
        X_P = _gen_delay_X(X, n_delay)
        separated_trainSig_1 = separate_trainSig(X_P, Y)
        separated_trainSig = [np.concatenate([trainSig_single_freq @ P_single_freq], axis=-1)
                              for trainSig_single_freq, P_single_freq in zip(separated_trainSig_1, ref_sig_P)]

        # template signals and spatial filters
        #   U: (filterbank_num * stimulus_num * channel_num * n_component)
        #   X or template_sig: (filterbank_num, channel_num, signal_len)
        filterbank_num, channel_num, signal_len = X_P[0].shape
        stimulus_num = len(ref_sig)
        n_component = self.n_component
        n_neighbor = min(self.n_neighbor,stimulus_num)
        d0 = int(np.floor(n_neighbor/2))
        _, freqs_idx, return_freqs_idx = sort(freqs)
        separated_trainSig = [separated_trainSig[i] for i in freqs_idx]
        separated_trainSig_1 = [separated_trainSig_1[i] for i in freqs_idx]
        # possible_class = list(set(Y))
        # possible_class.sort(reverse = False)
        # template_sig = [np.zeros((filterbank_num, channel_num, signal_len)) for _ in range(stimulus_num)]
        U_tdca = np.zeros((filterbank_num, 1, channel_num, stimulus_num))
        for filterbank_idx in range(filterbank_num):
            for class_idx in range(1,stimulus_num+1):
                if class_idx <= d0:
                    start_idx = 0
                    end_idx = n_neighbor
                elif class_idx > d0 and class_idx < (stimulus_num-d0+1):
                    start_idx = class_idx - d0 - 1
                    end_idx = class_idx + (n_neighbor-d0-1)
                else:
                    start_idx = stimulus_num - n_neighbor
                    end_idx = stimulus_num
                separated_trainSig_tmp = [separated_trainSig[i] for i in range(start_idx, end_idx)]
                # Calulcate spatial filter
                W = _TDCA_ls_calW([np.swapaxes(trainSig[:,filterbank_idx,:,:], 1,2) for trainSig in separated_trainSig_tmp],
                                self.n_jobs,
                                LSconfig = self.LSconfig)
                U_tdca[filterbank_idx,0,:,(class_idx-1):class_idx] = W[:,:n_component]
        # Calculate template
        separated_trainSig_1 = [separated_trainSig_1[i] for i in return_freqs_idx]
        template_sig = [np.mean(tmp, 0) for tmp in separated_trainSig_1]
        # Generate spatial filters for all stimuli
        # U_tdca = np.repeat(U_tdca[:,:,:,return_freqs_idx], repeats = stimulus_num, axis = 1)
        self.model['U'] = np.repeat(U_tdca, repeats = stimulus_num, axis = 1)
        self.model['template_sig'] = template_sig
    
    def predict(self,
            X: List[ndarray]) -> List[int]:
        weights_filterbank = self.model['weights_filterbank']
        if weights_filterbank is None:
            weights_filterbank = [1 for _ in range(X[0].shape[0])]
        if type(weights_filterbank) is list:
            weights_filterbank = np.expand_dims(np.array(weights_filterbank),1).T
        else:
            if len(weights_filterbank.shape) != 2:
                raise ValueError("'weights_filterbank' has wrong shape")
            if weights_filterbank.shape[0] != 1:
                weights_filterbank = weights_filterbank.T
        if weights_filterbank.shape[0] != 1:
            raise ValueError("'weights_filterbank' has wrong shape")
        n_delay = self.n_delay

        X_delay = _gen_delay_X(X, n_delay)

        template_sig = self.model['template_sig']
        U = self.model['U'] 
        ref_sig_P = self.model['ref_sig_P']

        if self.n_jobs is not None:
            r = Parallel(n_jobs=self.n_jobs)(delayed(partial(_r_tdca_canoncorr_withUV, Y=template_sig, P=ref_sig_P, U=U, V=U))(X=a) for a in X_delay)
        else:
            r = []
            for a in X_delay:
                r.append(
                    _r_tdca_canoncorr_withUV(X=a, Y=template_sig, P=ref_sig_P, U=U, V=U)
                )

        Y_pred = [int( np.argmax( weights_filterbank @ r_tmp)) for r_tmp in r]
        
        return Y_pred, r


# -*- coding: utf-8 -*-

from typing import Union, Optional, Dict, List, Tuple, Callable
from numpy import iscomplex, ndarray

import scipy.linalg as slin
import numpy as np
import numpy.linalg as nplin
import numpy.matlib as npmat

def get_depend_column_ind(X: ndarray):
    """
    Obtain depend column index

    Parameters
    ------------
    X : ndarray
        Input matrix
    """
    X_rank = nplin.matrix_rank(X)
    _, X_r = nplin.qr(X)
    X_r_diag = np.diag(X_r)
    X_r_diag_abssort_ind = np.argsort(np.abs(X_r_diag))[::-1]
    return X_r_diag_abssort_ind[X_rank:]


def remove_mean_all_trial(X : ndarray,
                          axis : int = 0, 
                          remove_val : Optional[ndarray] = None):
    """
    Remove mean of all trials

    Parameters
    ------------
    X : ndarray
        All trial signals
        Shape: [trials, samples, ch]
    axis : int
        Remove which axis's mean
    """
    Nt, Np, Nc = X.shape
    X_remove_mean = np.zeros_like(X)
    for i in range(Nt):
        if remove_val is None:
            X_remove_mean[i,:,:] = X[i,:,:] - np.mean(X[i,:,:], axis)
        else:
            X_remove_mean[i,:,:] = X[i,:,:] - remove_val
    return X_remove_mean

def column_eyes(number_of_eyes : int, 
                size_of_eyes : int):
    return repmat(np.eye(size_of_eyes),number_of_eyes,1)

def inv(X : ndarray):
    return nplin.inv(X)

def repmat(X : ndarray,
           rep_x : int,
           rep_y : int):
    return npmat.repmat(X, rep_x, rep_y)

def svd(X : ndarray,
        full_matrices : bool,
        compute_uv : bool):
    if compute_uv:
        L, D, M = slin.svd(X,
                            full_matrices=full_matrices,
                            compute_uv=compute_uv,
                            check_finite=False,
                            lapack_driver='gesvd')
        return L, D, M
    else:
        D = slin.svd(X,
                        full_matrices=full_matrices,
                        compute_uv=compute_uv,
                        check_finite=False,
                        lapack_driver='gesvd')
        return D

def cholesky(M : ndarray):
    """
    Calculate cholesky decomposition of M. If M is not positive definite matrix, the nearest positive definite matrix for M will be created.

    ref: https://github.com/Cysu/open-reid/commit/61f9c4a4da95d0afc3634180eee3b65e38c54a14

    Find the nearest positive definite matrix for M. Modified from
    http://www.mathworks.com/matlabcentral/fileexchange/42885-nearestspd
    Might take several minutes
    """
    M = (M + M.T) * 0.5
    k = 0
    I = np.eye(M.shape[0])
    Ki = None
    while k<=1000:
        try:
            Ki = nplin.cholesky(M)
            break
        except nplin.LinAlgError:
            k += 1
            v = eigvec(M)
            min_eig = v.min()
            M += (-min_eig * k * k + np.spacing(min_eig)) * I
    if Ki is None:
        raise ValueError('Cannot calculate cholesky decomposition')
    else:
        return Ki

def norm_direction(V : ndarray,
                   V_norm : Optional[List[float]] = None):
    """
    Normalize directions of colums 
    """
    _, n_col = V.shape
    if V_norm is None:
        if n_col == 1:
            V_norm = [1]
        else:
            V_norm = []
            for col_idx in range(n_col):
                v1 = V[:,0]/nplin.norm(V[:,0])
                v2 = V[:,col_idx]/nplin.norm(V[:,col_idx])
                v_dot = np.dot(v1, v2)
                if v_dot > 1:
                    v_dot = 1
                if v_dot < -1:
                    v_dot = -1
                v_angle = np.arccos(v_dot)/np.pi
                if v_angle > 0.5:
                    V[:,col_idx] = -1 * V[:,col_idx]
                    V_norm.append(-1)
                else:
                    V_norm.append(1)
    else:
        for col_idx in range(n_col):
            V[:,col_idx] = V_norm[col_idx] * V[:,col_idx]
    return V, V_norm

def eigvec(X : ndarray,
           Y : Optional[ndarray] = None):
    """
    Calculate eigenvectors

    Parameters
    -----------------
    X : ndarray
        A complex or real matrix whose eigenvalues and eigenvectors will be computed.
    Y : ndarray
        If Y is given, eig(Y\X), or say  eig(X, Y), will be computed

    Returns
    ---------------
    eig_vec : ndarray
        Eigenvectors. The order follows the corresponding eigenvalues (from high to low values)
    """
    if Y is None:
        eig_d1, eig_v1 = slin.eig(X) #eig(X)
    else:
        eig_d1, eig_v1 = slin.eig(X, Y) #eig(Y\X)

    if len(eig_d1.shape) == 2:
        eig_d1 = np.diagonal(eig_d1)

    sort_idx = np.argsort(eig_d1)[::-1]
    eig_vec = eig_v1[:,sort_idx]

    if Y is not None:
        square_val = np.diag(eig_vec.T @ Y @ eig_vec)
        norm_v = np.sqrt(square_val)
        eig_vec = eig_vec/norm_v

    if np.iscomplex(eig_vec).any():
        eig_vec = np.real(eig_vec)

    return eig_vec

def sum_list(X: list) -> ndarray:
    """
    Calculate sum of a list

    Parameters
    ------------
    X: list

    Returns
    -------------
    sum_X: ndarray
    """
    sum_X = None
    for x in X:
        if type(x) is list:
            x = sum_list(x)
        if sum_X is None:
            sum_X = x
        else:
            sum_X = sum_X + x
    return sum_X

def mean_list(X: list) -> ndarray:
    """
    Calculate mean of a list

    Parameters
    -----------
    X: list

    Returns
    ----------
    mean_X: ndarray
    """
    tmp = []
    for X_single_trial in X:
        if type(X_single_trial) is list:
            X_single_trial = mean_list(X_single_trial)
        tmp.append(np.expand_dims(X_single_trial, axis = 0))
    tmp = np.concatenate(tmp, axis = 0)
    return np.mean(tmp, axis=0)

def blkmat(X: ndarray):
    """
    Build the block diag matrix by using X

    Parameters
    -----------
    X : ndarray or list
        Matrix used for building the block diag matrix
        (trial_num, channel_num, signal_len) or [(channel_num, signal_len)]
    """
    blkmatrix = np.ndarray([])
    if type(X) == np.ndarray:
        trial_num, channel_num, signal_len = X.shape
        for trial_idx in range(trial_num):
            if len(blkmatrix.shape)==0:
                blkmatrix = X[trial_idx,:,:]
            else:
            #     A1 = np.concatenate((blkmatrix, np.zeros((blkmatrix.shape[0], signal_len))), axis = 1)
            #     A2 = np.concatenate((np.zeros((channel_num, blkmatrix.shape[1])), X[trial_idx,:,:]), axis = 1)
            #     blkmatrix = np.concatenate((A1, A2), axis = 0)
                blkmatrix = slin.block_diag(blkmatrix, X[trial_idx,:,:])
    elif type(X) == list:
        for tmp in X:
            if len(blkmatrix.shape)==0:
                blkmatrix = tmp
            else:
                blkmatrix = slin.block_diag(blkmatrix, tmp)
    else:
        raise ValueError('Unknown input type in blkmat')
    return blkmatrix

def blkrep(X: ndarray,
           N: int):
    """
    Build the block diag matrix by repeat X with N times

    Parameters
    -----------
    X : ndarray
        Matrix used for building the block diag matrix
    N : int
        Number of X in the diag line
    """
    blkmatrix = np.ndarray([])
    for _ in range(N):
        if len(blkmatrix.shape)==0:
            blkmatrix = X.copy()
        else:
        #     A1 = np.concatenate((blkmatrix, np.zeros((blkmatrix.shape[0], X.shape[1]))), axis = 1)
        #     A2 = np.concatenate((np.zeros((X.shape[0], blkmatrix.shape[1])), X), axis = 1)
        #     blkmatrix = np.concatenate((A1, A2), axis = 0)
            blkmatrix = slin.block_diag(blkmatrix, X)
    return blkmatrix


def sort(X: list) -> Tuple[list, list, list]:
    """
    Sort given list

    Parameters
    -----------
    X : list

    Returns
    -----------
    sorted_X : list
        Sorted X
    sort_idx : list
        Indices that is applied to transfer list from X to sorted_X
    return_idx : list
        Indices that is applied to transfer list from sorted_X to X
    """
    sort_idx = list(np.argsort(X))
    sorted_X = [X[i] for i in sort_idx]
    return_idx = [None] * len(sort_idx)
    for loc, idx in enumerate(sort_idx):
        return_idx[idx] = loc
    return sorted_X, sort_idx, return_idx

def separate_trainSig(X: List[ndarray],
                      Y: List[int]) -> List[ndarray]:
    """
    Separate training signals

    Parameters
    ----------
    X : List[ndarray]
        Training data
        List shape: (trial_num,)
        EEG shape: (filterbank_num, channel_num, signal_len)
    Y : List[int]
        Training label
        List shape: (trial_num,)

    Returns
    -------
    template_sig : List[ndarray]
        Template signal
        List of shape: (stimulus_num,)
        Template shape: (trial_num, filterbank_num, channel_num, signal_len)
    """
    # Get possible stimulus class
    unique_Y = list(set(Y))
    unique_Y.sort()
    # 
    template_sig = []
    for i in unique_Y:
        # i-th class trial index
        target_idx = [k for k in range(len(Y)) if Y[k] == unique_Y[i]]
        # Get i-th class training data
        template_sig_single = [np.expand_dims(X[k], axis=0) for k in target_idx]
        template_sig_single = np.concatenate(template_sig_single, axis=0)
        # Store i-th class template
        template_sig.append(template_sig_single)
    return template_sig

def gen_template(X: List[ndarray],
                 Y: List[int]) -> List[ndarray]:
    """
    Average training data according to training label to generate template signal

    Parameters
    ----------
    X : List[ndarray]
        Training data
        List shape: (trial_num,)
        EEG shape: (filterbank_num, channel_num, signal_len)
    Y : List[int]
        Training label
        List shape: (trial_num,)

    Returns
    -------
    template_sig : List[ndarray]
        Template signal
        List of shape: (stimulus_num,)
        Template shape: (filterbank_num, channel_num, signal_len)
    """
    # Get possible stimulus class
    unique_Y = list(set(Y))
    unique_Y.sort()
    # 
    template_sig = []
    for i in unique_Y:
        # i-th class trial index
        target_idx = [k for k in range(len(Y)) if Y[k] == unique_Y[i]]
        # Get i-th class training data
        template_sig_single = [np.expand_dims(X[k], axis=0) for k in target_idx]
        template_sig_single = np.concatenate(template_sig_single, axis=0)
        # Average all i-th class training data
        template_sig_single = np.mean(template_sig_single, axis=0)
        # Store i-th class template
        template_sig.append(template_sig_single)
    return template_sig

def canoncorr(X: ndarray, 
              Y: ndarray,
              force_output_UV: Optional[bool] = False) -> Union[Tuple[ndarray, ndarray, ndarray], ndarray]:
    """
    Canonical correlation analysis following matlab

    Parameters
    ----------
    X : ndarray
    Y : ndarray
    force_output_UV : Optional[bool]
        whether calculate and output A and B
    
    Returns
    -------
    A : ndarray
        if force_output_UV, return A
    B : ndarray
        if force_output_UV, return B
    r : ndarray
    """
    n, p1 = X.shape
    _, p2 = Y.shape
    
    Q1, T11, perm1 = qr_remove_mean(X)
    Q2, T22, perm2 = qr_remove_mean(Y)
    
    svd_X = Q1.T @ Q2
    if svd_X.shape[0]>svd_X.shape[1]:
        full_matrices=False
    else:
        full_matrices=True
        
    L, D, M = slin.svd(svd_X,
                     full_matrices=full_matrices,
                     check_finite=False,
                     lapack_driver='gesvd')
    M = M.T
    
    r = D
    
    if force_output_UV:
        A = mldivide(T11, L) * np.sqrt(n - 1)
        B = mldivide(T22, M) * np.sqrt(n - 1)
        A_r = np.zeros(A.shape)
        for i in range(A.shape[0]):
            A_r[perm1[i],:] = A[i,:]
        B_r = np.zeros(B.shape)
        for i in range(B.shape[0]):
            B_r[perm2[i],:] = B[i,:]
            
        return A_r, B_r, r
    else:
        return r

def qr_inverse(Q: ndarray, 
               R: ndarray,
               P: ndarray) -> ndarray:
    """
    Inverse QR decomposition

    Parameters
    ----------
    Q : ndarray
        (M * K) - reference
        (filterbank_num * M * K) - template
    R : ndarray
        (K * N) - reference
        (filterbank_num * K * N) - template
    P : ndarray
        (N,) - reference
        (filterbank_num * N) - template

    Returns
    -------
    X : ndarray
        (M * N) - reference
        (filterbank_num * M * N) - template
    """
    if len(Q.shape)==2: # reference
        tmp = Q @ R
        X = np.zeros(tmp.shape)
        for i in range(X.shape[1]):
            X[:,P[i]] = tmp[:,i]
    elif len(Q.shape)==3: # template
        X = [np.expand_dims(qr_inverse(Q[i,:,:], R[i,:,:], P[i,:]), axis=0) for i in range(Q.shape[0])]
        X = np.concatenate(X, axis=0)
    else:
        raise ValueError('Unknown data type')
    return X

def qr_list(X : List[ndarray]) -> Tuple[List[ndarray], List[ndarray], List[ndarray]]:
    """
    QR decomposition of list X
    Note: Elements in X will be transposed first and then decomposed

    Parameters
    ----------
    X : List[ndarray]

    Returns
    -------
    Q : List[ndarray]
    R : List[ndarray]
    P : List[ndarray]
    """
    Q = []
    R = []
    P = []
    for el in X:
        if len(el.shape) == 2: # reference signal
            Q_tmp, R_tmp, P_tmp = qr_remove_mean(el.T)
            Q.append(Q_tmp)
            R.append(R_tmp)
            P.append(P_tmp)
        elif len(el.shape) == 3: # template signal
            Q_tmp = []
            R_tmp = []
            P_tmp = []
            for k in range(el.shape[0]):
                Q_tmp_tmp, R_tmp_tmp, P_tmp_tmp = qr_remove_mean(el[k,:,:].T)
                Q_tmp.append(np.expand_dims(Q_tmp_tmp, axis=0))
                R_tmp.append(np.expand_dims(R_tmp_tmp, axis=0))
                P_tmp.append(np.expand_dims(P_tmp_tmp, axis=0))
            Q.append(np.concatenate(Q_tmp,axis=0))
            R.append(np.concatenate(R_tmp,axis=0))
            P.append(np.concatenate(P_tmp,axis=0))
        else:
            raise ValueError('Unknown data type')
    return Q, R, P

def qr_remove_mean(X: ndarray) -> Tuple[ndarray, ndarray, ndarray]:
    """
    Remove column mean and QR decomposition 

    Parameters
    ----------
    X : ndarray
        (M * N)

    Returns
    -------
    Q : ndarray
        (M * K)
    R : ndarray
        (K * N)
    P : ndarray
        (N,)
    """
    
    X_remove_mean = X - np.mean(X,0)
    
    Q, R, P = slin.qr(X_remove_mean, mode = 'economic', pivoting = True)
    
    return Q, R, P

def mldivide(A: ndarray,
             B: ndarray) -> ndarray:
    """
    A\B, Solve Ax = B

    Parameters
    ----------
    A : ndarray
    B : ndarray

    Returns
    -------
    x: ndarray
    """
    
    return slin.pinv(A) @ B